r"""
Universal functions.

Ufuncs have class-oriented and function-oriented interfaces.
Only the functional interface is exposed due to its common use.

Example
-------
.. code-block:: python3

   from pyxu.abc import LinOp
   from pyxu.operator import sin
   from pyxu.operator.map.ufunc import _Sin

   N = 10
   x = np.random.randn(N)
   A = LinOp.from_array(np.random.randn(N, N))

   op1 = _Sin(x.size) * A  # class interface
   op2 = sin(A)           # function interface

   np.allclose(op1.apply(x), np.sin(A.apply(x)))  # True
   np.allclose(op2.apply(x), np.sin(A.apply(x)))  # True
"""

import numpy as np

import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.operator.linop.base as pxlb
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "sin",
    "cos",
    "tan",
    "arcsin",
    "arccos",
    "arctan",
    "sinh",
    "cosh",
    "tanh",
    "arcsinh",
    "arccosh",
    "arctanh",
    "exp",
    "log",
    "clip",
    "sqrt",
    "cbrt",
    "square",
    "abs",
    "sign",
    "gaussian",
    "sigmoid",
    "softplus",
    "leakyrelu",
    "relu",
    "silu",
    "softmax",
]


# Trigonometric Functions =====================================================
class _Sin(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = 1
        self.diff_lipschitz = 1

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.sin(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        xp = pxu.get_array_module(arr)
        return pxlb.DiagonalOp(xp.cos(arr))


def sin(op: pxt.OpT) -> pxt.OpT:
    r"""
    Trigonometric sine, element-wise.

    Notes
    -----
    * :math:`f(x) = \sin(x)`
    * :math:`f'(x) = \cos(x)`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 1`.

      (Reason: :math:`\vert f'(x) \vert` is bounded by :math:`L` at :math:`x = k \pi, \, k \in \mathbb{Z}`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = 1`.

      (Reason: :math:`\vert f''(x) \vert` is bounded by :math:`\partial L` at :math:`x = (2k + 1) \frac{\pi}{2}, \, k
      \in \mathbb{Z}`.)
    """
    return _Sin(op.dim) * op


class _Cos(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = 1
        self.diff_lipschitz = 1

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.cos(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        xp = pxu.get_array_module(arr)
        return pxlb.DiagonalOp(-xp.sin(arr))


def cos(op: pxt.OpT) -> pxt.OpT:
    r"""
    Trigonometric cosine, element-wise.

    Notes
    -----
    * :math:`f(x) = \cos(x)`
    * :math:`f'(x) = -\sin(x)`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 1`.

      (Reason: :math:`\vert f'(x) \vert` is bounded by :math:`L` at :math:`x = (2k + 1) \frac{\pi}{2}, \, k \in
      \mathbb{Z}`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = 1`.

      (Reason: :math:`\vert f''(x) \vert` is bounded by :math:`\partial L` at :math:`x = k \pi, \, k \in \mathbb{Z}`.)
    """
    return _Cos(op.dim) * op


class _Tan(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = np.inf
        self.diff_lipschitz = np.inf

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.tan(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        xp = pxu.get_array_module(arr)
        v = xp.cos(arr)
        v **= 2
        return pxlb.DiagonalOp(1 / v)


def tan(op: pxt.OpT) -> pxt.OpT:
    r"""
    Trigonometric tangent, element-wise.

    Notes
    -----
    * :math:`f(x) = \tan(x)`
    * :math:`f'(x) = \cos^{-2}(x)`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \infty`.

      (Reason: :math:`f'(x)` is unbounded on :math:`\text{dom}(f) = [-\pi, \pi]`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = \infty`.

      (Reason: :math:`f''(x)` is unbounded on :math:`\text{dom}(f) = [-\pi, \pi]`.)
    """
    return _Tan(op.dim) * op


class _ArcSin(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = np.inf
        self.diff_lipschitz = np.inf

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.arcsin(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        xp = pxu.get_array_module(arr)
        v = arr**2
        v *= -1
        v += 1
        xp.sqrt(v, out=v)
        return pxlb.DiagonalOp(1 / v)


def arcsin(op: pxt.OpT) -> pxt.OpT:
    r"""
    Inverse sine, element-wise.

    Notes
    -----
    * :math:`f(x) = \arcsin(x)`
    * :math:`f'(x) = (1 - x^{2})^{-\frac{1}{2}}`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \infty`.

      (Reason: :math:`f'(x)` is unbounded on :math:`\text{dom}(f) = [-1, 1]`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = \infty`.

      (Reason: :math:`f''(x)` is unbounded on :math:`\text{dom}(f) = [-1, 1]`.)
    """
    return _ArcSin(op.dim) * op


class _ArcCos(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.arccos(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        xp = pxu.get_array_module(arr)
        v = arr**2
        v *= -1
        v += 1
        xp.sqrt(v, out=v)
        return pxlb.DiagonalOp(-1 / v)


def arccos(op: pxt.OpT) -> pxt.OpT:
    r"""
    Inverse cosine, element-wise.

    Notes
    -----
    * :math:`f(x) = \arccos(x)`
    * :math:`f'(x) = -(1 - x^{2})^{-\frac{1}{2}}`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \infty`.

      (Reason: :math:`f'(x)` is unbounded on :math:`\text{dom}(f) = [-1, 1]`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = \infty`.

      (Reason: :math:`f''(x)` is unbounded on :math:`\text{dom}(f) = [-1, 1]`.)
    """
    return _ArcCos(op.dim) * op


class _ArcTan(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = 1
        self.diff_lipschitz = 3 * np.sqrt(3) / 8
        #   max_{x \in R} |arctan''(x)|
        # = max_{x \in R} |2x / (1+x^2)^2|
        # = 3 \sqrt(3) / 8  [at x = +- 1/\sqrt(3)]

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.arctan(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        v = arr**2
        v += 1
        return pxlb.DiagonalOp(1 / v)


def arctan(op: pxt.OpT) -> pxt.OpT:
    r"""
    Inverse tangent, element-wise.

    Notes
    -----
    * :math:`f(x) = \arctan(x)`
    * :math:`f'(x) = (1 + x^{2})^{-1}`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 1`.

      (Reason: :math:`\vert f'(x) \vert` is bounded by :math:`L` at :math:`x = 0`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = 3 \sqrt{3} / 8`.

      (Reason: :math:`\vert f''(x) \vert` is bounded by :math:`\partial L` at :math:`x = \pm \frac{1}{\sqrt{3}}`.)
    """
    return _ArcTan(op.dim) * op


# Hyperbolic Functions ========================================================
class _Sinh(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = np.inf
        self.diff_lipschitz = np.inf

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.sinh(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        xp = pxu.get_array_module(arr)
        return pxlb.DiagonalOp(xp.cosh(arr))


def sinh(op: pxt.OpT) -> pxt.OpT:
    r"""
    Hyperbolic sine, element-wise.

    Notes
    -----
    * :math:`f(x) = \sinh(x)`
    * :math:`f'(x) = \cosh(x)`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \infty`.

      (Reason: :math:`f'(x)` is unbounded on :math:`\text{dom}(f) = \mathbb{R}`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = \infty`.

      (Reason: :math:`f''(x)` is unbounded on :math:`\text{dom}(f) = \mathbb{R}`.)
    """
    return _Sinh(op.dim) * op


class _Cosh(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = np.inf
        self.diff_lipschitz = np.inf

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.cosh(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        xp = pxu.get_array_module(arr)
        return pxlb.DiagonalOp(xp.sinh(arr))


def cosh(op: pxt.OpT) -> pxt.OpT:
    r"""
    Hyperbolic cosine, element-wise.

    Notes
    -----
    * :math:`f(x) = \cosh(x)`
    * :math:`f'(x) = \sinh(x)`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \infty`.

      (Reason: :math:`f'(x)` is unbounded on :math:`\text{dom}(f) = \mathbb{R}`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = \infty`.

      (Reason: :math:`f''(x)` is unbounded on :math:`\text{dom}(f) = \mathbb{R}`.)
    """
    return _Cosh(op.dim) * op


class _Tanh(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = 1
        self.diff_lipschitz = 4 / (3 * np.sqrt(3))
        #   max_{x \in R} |tanh''(x)|
        # = max_{x \in R} |-2 tanh(x) [1 - tanh(x)^2|
        # = 4 / (3 \sqrt(3))  [at x = ln(2 +- \sqrt(3))]

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.tanh(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        v = self.apply(arr)
        v**2
        v *= -1
        v += 1
        return pxlb.DiagonalOp(v)


def tanh(op: pxt.OpT) -> pxt.OpT:
    r"""
    Hyperbolic tangent, element-wise.

    Notes
    -----
    * :math:`f(x) = \tanh(x)`
    * :math:`f'(x) = 1 - \tanh^{2}(x)`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 1`.

      (Reason: :math:`\vert f'(x) \vert` is bounded by :math:`L` at :math:`x = 0`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = 4 / 3 \sqrt{3}`.

      (Reason: :math:`\vert f''(x) \vert` is bounded by :math:`\partial L` at :math:`x = \frac{1}{2} \ln(2 \pm
      \sqrt{3})`.
    """
    return _Tanh(op.dim) * op


class _ArcSinh(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = 1
        self.diff_lipschitz = 2 / (3 * np.sqrt(3))
        #   max_{x \in R} |arcsinh''(x)|
        # = max_{x \in R} |-x (x^2 + 1)^{-3/2}|
        # = 2 / (3 \sqrt(3))  [at x = += 1 / \sqrt(2)]

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.arcsinh(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        xp = pxu.get_array_module(arr)
        v = arr**2
        v += 1
        xp.sqrt(v, out=v)
        return pxlb.DiagonalOp(1 / v)


def arcsinh(op: pxt.OpT) -> pxt.OpT:
    r"""
    Inverse hyperbolic sine, element-wise.

    Notes
    -----
    * :math:`f(x) = \sinh^{-1}(x) = \ln(x + \sqrt{x^{2} + 1})`
    * :math:`f'(x) = (x^{2} + 1)^{-\frac{1}{2}}`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 1`.

      (Reason: :math:`\vert f'(x) \vert` is bounded by :math:`L` at :math:`x = 0`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = \frac{2}{3 \sqrt{3}}`.

      (Reason: :math:`\vert f''(x) \vert` is bounded by :math:`\partial L` at :math:`x = \pm \frac{1}{\sqrt{2}}`.)
    """
    return _ArcSinh(op.dim) * op


class _ArcCosh(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = np.inf
        self.diff_lipschitz = np.inf

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.arccosh(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        xp = pxu.get_array_module(arr)
        v = arr**2
        v -= 1
        xp.sqrt(v, out=v)
        return pxlb.DiagonalOp(1 / v)


def arccosh(op: pxt.OpT) -> pxt.OpT:
    r"""
    Inverse hyperbolic cosine, element-wise.

    Notes
    -----
    * :math:`f(x) = \cosh^{-1}(x) = \ln(x + \sqrt{x^{2} - 1})`
    * :math:`f'(x) = (x^{2} - 1)^{-\frac{1}{2}}`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \infty`.

      (Reason: :math:`f'(x)` is unbounded on :math:`\text{dom}(f) = [1, \infty[`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = \infty`.

      (Reason: :math:`f''(x)` is unbounded on :math:`\text{dom}(f) = [1, \infty[`.)
    """
    return _ArcCosh(op.dim) * op


class _ArcTanh(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = np.inf
        self.diff_lipschitz = np.inf

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.arctanh(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        v = arr**2
        v *= -1
        v += 1
        return pxlb.DiagonalOp(1 / v)


def arctanh(op: pxt.OpT) -> pxt.OpT:
    r"""
    Inverse hyperbolic tangent, element-wise.

    Notes
    -----
    * :math:`f(x) = \tanh^{-1}(x) = \frac{1}{2}\ln\left(\frac{1+x}{1-x}\right)`
    * :math:`f'(x) = (1 - x^{2})^{-1}`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \infty`.

      (Reason: :math:`f'(x)` is unbounded on :math:`\text{dom}(f) = [-1, 1]`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = \infty`.

      (Reason: :math:`f''(x)` is unbounded on :math:`\text{dom}(f) = [-1, 1]`.)
    """
    return _ArcTanh(op.dim) * op


# Exponential Functions =======================================================
class _Exp(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer, base: pxt.Real = None):
        super().__init__(shape=(dim, dim))
        self.lipschitz = np.inf
        self.diff_lipschitz = np.inf
        self._base = base

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        out = arr.copy()
        if self._base is not None:
            out *= np.log(float(self._base))

        xp = pxu.get_array_module(arr)
        xp.exp(out, out=out)
        return out

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        v = self.apply(arr)
        if self._base is not None:
            v *= np.log(float(self._base))
        return pxlb.DiagonalOp(v)


def exp(op: pxt.OpT, base: pxt.Real = None) -> pxt.OpT:
    r"""
    Exponential, element-wise. (Default: base-E exponential.)

    Notes
    -----
    * :math:`f_{b}(x) = b^{x}`
    * :math:`f_{b}'(x) = b^{x} \ln(b)`
    * :math:`\vert f_{b}(x) - f_{b}(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \infty`.

      (Reason: :math:`f_{b}'(x)` is unbounded on :math:`\text{dom}(f_{b}) = \mathbb{R}`.)
    * :math:`\vert f_{b}'(x) - f_{b}'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant
      :math:`\partial L = \infty`.

      (Reason: :math:`f_{b}''(x)` is unbounded on :math:`\text{dom}(f_{b}) = \mathbb{R}`.)
    """
    return _Exp(op.dim, base) * op


class _Log(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer, base: pxt.Real = None):
        super().__init__(shape=(dim, dim))
        self.lipschitz = np.inf
        self.diff_lipschitz = np.inf
        self._base = base

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        out = xp.log(arr)
        if self._base is not None:
            out /= np.log(float(self._base))
        return out

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        v = 1 / arr
        if self._base is not None:
            v /= np.log(float(self._base))
        return pxlb.DiagonalOp(v)


def log(op: pxt.OpT, base: pxt.Real = None) -> pxt.OpT:
    r"""
    Logarithm, element-wise. (Default: base-E logarithm.)

    Notes
    -----
    * :math:`f_{b}(x) = \log_{b}(x)`
    * :math:`f_{b}'(x) = x^{-1} / \ln(b)`
    * :math:`\vert f_{b}(x) - f_{b}(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \infty`.

      (Reason: :math:`f_{b}'(x)` is unbounded on :math:`\text{dom}(f_{b}) = \mathbb{R}_{+}`.)
    * :math:`\vert f_{b}'(x) - f_{b}'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant
      :math:`\partial L = \infty`.

      (Reason: :math:`f_{b}''(x)` is unbounded on :math:`\text{dom}(f_{b}) = \mathbb{R}_{+}`.)
    """
    return _Log(op.dim, base) * op


# Miscellaneous ===============================================================
class _Clip(pxa.Map):
    def __init__(self, dim: pxt.Integer, a_min: pxt.Real = None, a_max: pxt.Real = None):
        super().__init__(shape=(dim, dim))
        if (a_min is None) and (a_max is None):
            raise ValueError("One of Parameter[a_min, a_max] must be specified.")
        self._llim = a_min
        self._ulim = a_max
        self.lipschitz = 1

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        out = arr.copy()
        xp.clip(
            arr,
            self._llim,
            self._ulim,
            out=out,
        )
        return out


def clip(op: pxt.OpT, a_min: pxt.Real = None, a_max: pxt.Real = None):
    r"""
    Clip (limit) values in an array, element-wise.

    Notes
    -----
    * .. math::

         f_{[a,b]}(x) =
         \begin{cases}
             a, & \text{if} \ x \leq a, \\
             x, & a < x < b, \\
             b, & \text{if} \ x \geq b.
         \end{cases}
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 1`.
    """
    return _Clip(op.dim, a_min=a_min, a_max=a_max) * op


class _Sqrt(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = np.inf
        self.diff_lipschitz = np.inf

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.sqrt(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        v = self.apply(arr)
        v *= 2
        return pxlb.DiagonalOp(1 / v)


def sqrt(op: pxt.OpT) -> pxt.OpT:
    r"""
    Non-negative square-root, element-wise.

    Notes
    -----
    * :math:`f(x) = \sqrt{x}`
    * :math:`f'(x) = 1 / 2 \sqrt{x}`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \infty`.

      (Reason: :math:`f'(x)` is unbounded on :math:`\text{dom}(f) = \mathbb{R}_{+}`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = \infty`.

      (Reason: :math:`f''(x)` is unbounded on :math:`\text{dom}(f) = \mathbb{R}_{+}`.)
    """
    return _Sqrt(op.dim) * op


class _Cbrt(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = np.inf
        self.diff_lipschitz = np.inf

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.cbrt(arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        v = self.apply(arr)
        v **= 2
        v *= 3
        return pxlb.DiagonalOp(1 / v)


def cbrt(op: pxt.OpT) -> pxt.OpT:
    r"""
    Cube-root, element-wise.

    Notes
    -----
    * :math:`f(x) = \sqrt[3]{x}`
    * :math:`f'(x) = 1 / 3 \sqrt[3]{x^{2}}`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \infty`.

      (Reason: :math:`f'(x)` is unbounded on :math:`\text{dom}(f) = \mathbb{R}`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = \infty`.

      (Reason: :math:`f''(x)` is unbounded on :math:`\text{dom}(f) = \mathbb{R}`.)
    """
    return _Cbrt(op.dim) * op


class _Square(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = np.inf
        self.diff_lipschitz = 2

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        out = arr.copy()
        out **= 2
        return out

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        v = arr.copy()
        v *= 2
        return pxlb.DiagonalOp(v)


def square(op: pxt.OpT) -> pxt.OpT:
    r"""
    Square, element-wise.

    Notes
    -----
    * :math:`f(x) = x^{2}`
    * :math:`f'(x) = 2 x`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \infty`.

      (Reason: :math:`f'(x)` is unbounded on :math:`\text{dom}(f) = \mathbb{R}`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = 2`.

      (Reason: :math:`\vert f''(x) \vert` is bounded by :math:`\partial L` everywhere.)
    """
    return _Square(op.dim) * op


class _Abs(pxa.Map):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = 1

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.abs(arr)


def abs(op: pxt.OpT) -> pxt.OpT:
    r"""
    Absolute value, element-wise.

    Notes
    -----
    * :math:`f(x) = \vert x \vert`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 1`.
    """
    return _Abs(op.dim) * op


class _Sign(pxa.Map):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = 2

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.sign(arr)


def sign(op: pxt.OpT) -> pxt.OpT:
    r"""
    Number sign indicator, element-wise.

    Notes
    -----
    * :math:`f(x) = x / \vert x \vert`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 2`.
    """
    return _Sign(op.dim) * op


# Activation Functions ========================================================
class _Gaussian(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = np.sqrt(2 / np.e)
        self.diff_lipschitz = 2

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        out = arr.copy()
        out **= 2
        out *= -1
        xp.exp(out, out=out)
        return out

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        v = self.apply(arr)
        v *= -2
        v *= arr
        return pxlb.DiagonalOp(v)


def gaussian(op: pxt.OpT) -> pxt.OpT:
    r"""
    Gaussian, element-wise.

    Notes
    -----
    * :math:`f(x) = \exp(-x^{2})`
    * :math:`f'(x) = -2 x \exp(-x^{2})`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \sqrt{2 / e}`.

      (Reason: :math:`\vert f'(x) \vert` is bounded by :math:`L` at :math:`x = \pm 1 / \sqrt{2}`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = 2`.

      (Reason: :math:`\vert f''(x) \vert` is bounded by :math:`\partial L` at :math:`x = 0`.)
    """
    return _Gaussian(op.dim) * op


class _Sigmoid(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = 1 / 4
        self.diff_lipschitz = 1 / (6 * np.sqrt(3))

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        x = -arr
        xp.exp(x, out=x)
        x += 1
        return 1 / x

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        x = self.apply(arr)
        v = x.copy()
        x -= 1
        v *= x
        return pxlb.DiagonalOp(v)


def sigmoid(op: pxt.OpT) -> pxt.OpT:
    r"""
    Sigmoid, element-wise.

    Notes
    -----
    * :math:`f(x) = (1 + e^{-x})^{-1}`
    * :math:`f'(x) = f(x) [ f(x) - 1 ]`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 1 / 4`.

      (Reason: :math:`\vert f'(x) \vert` is bounded by :math:`L` at :math:`x = 0`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = 1 / 6 \sqrt{3}`.

      (Reason: :math:`\vert f''(x) \vert` is bounded by :math:`\partial L` at :math:`x = \ln(2 \pm \sqrt{3})`.)
    """
    return _Sigmoid(op.dim) * op


class _SoftPlus(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = 1
        self.diff_lipschitz = 1 / 4

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.logaddexp(0, arr)

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        f = _Sigmoid(dim=self.dim)
        v = f.apply(arr)
        return pxlb.DiagonalOp(v)


def softplus(op: pxt.OpT) -> pxt.OpT:
    r"""
    Softplus operator.

    Notes
    -----
    * :math:`f(x) = \ln(1 + e^{x})`
    * :math:`f'(x) = (1 + e^{-x})^{-1}`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 1`.

      (Reason: :math:`\vert f'(x) \vert` is bounded by :math:`L` on :math:`\text{dom}(f) = \mathbb{R}`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = 1 / 4`.

      (Reason: :math:`\vert f''(x) \vert` is bounded by :math:`\partial L` at :math:`x = 0`.)
    """
    return _SoftPlus(op.dim) * op


class _LeakyReLU(pxa.Map):
    def __init__(self, dim: pxt.Integer, alpha: pxt.Real):
        super().__init__(shape=(dim, dim))
        self._alpha = float(alpha)
        assert self._alpha >= 0
        self.lipschitz = float(max(alpha, 1))

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.where(arr >= 0, arr, arr * self._alpha)


def leakyrelu(op: pxt.OpT, alpha: pxt.Real) -> pxt.OpT:
    r"""
    Leaky rectified linear unit, element-wise.

    Notes
    -----
    * :math:`f(x) = x \left[\mathbb{1}_{\ge 0}(x) + \alpha \mathbb{1}_{< 0}(x)\right], \quad \alpha \ge 0`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = \max(1, \alpha)`.
    """
    return _LeakyReLU(dim=op.dim, alpha=alpha) * op


class _ReLU(_LeakyReLU):
    def __init__(self, dim: pxt.Integer):
        super().__init__(dim=dim, alpha=0)


def relu(op: pxt.OpT) -> pxt.OpT:
    r"""
    Rectified linear unit, element-wise.

    Notes
    -----
    * :math:`f(x) = \lfloor x \rfloor_{+}`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 1`.
    """
    return _ReLU(op.dim) * op


class _SiLU(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = 1.1
        self.diff_lipschitz = 1 / 2

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        f = _Sigmoid(dim=self.dim)
        out = f.apply(arr)
        out *= arr
        return out

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        f = _Sigmoid(dim=self.dim)
        xp = pxu.get_array_module(arr)
        a = xp.exp(-arr)
        a *= 1 + arr
        a += 1
        b = f.apply(arr)
        b **= 2
        return pxlb.DiagonalOp(a * b)


def silu(op: pxt.OpT) -> pxt.OpT:
    r"""
    Sigmoid linear unit, element-wise.

    Notes
    -----
    * :math:`f(x) = x / (1 + e^{-x})`
    * :math:`f'(x) = (1 + e^{-x} + x e^{-x}) / (1 + e^{-x})^{2}`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 1.1`.

      (Reason: :math:`\vert f'(x) \vert` is bounded by :math:`L` at :math:`x \approx 2.4`.)
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = 1 / 2`.

      (Reason: :math:`\vert f''(x) \vert` is bounded by :math:`\partial L` at :math:`x = 0`.)
    """
    return _SiLU(op.dim) * op


class _SoftMax(pxa.DiffMap):
    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(dim, dim))
        self.lipschitz = 1
        self.diff_lipschitz = 1

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        x = xp.exp(arr)
        out = x / x.sum(axis=-1, keepdims=True)
        return out

    @pxrt.enforce_precision(i="arr", o=False)
    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        v = self.apply(arr)
        lhs = pxlb.DiagonalOp(v)
        rhs = pxa.LinFunc.from_array(v).gram()
        op = lhs - rhs
        return op


def softmax(op: pxt.OpT) -> pxt.OpT:
    r"""
    Softmax operator.

    Notes
    -----
    * :math:`[f(x_{1},\ldots,x_{N})]_{i} = e^{x_{i}} / \sum_{k=1}^{N} e^{x_{k}}`
    * :math:`J_{f}(\mathbf{x}) = \text{diag}(f(\mathbf{x})) - f(\mathbf{x}) f(\mathbf{x})^{T}`
    * :math:`\vert f(x) - f(y) \vert \le L \vert x - y \vert`, with Lipschitz constant :math:`L = 1`.
    * :math:`\vert f'(x) - f'(y) \vert \le \partial L \vert x - y \vert`, with diff-Lipschitz constant :math:`\partial L
      = 1`.
    """
    return _SoftMax(op.dim) * op



import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class ConstantValueMixin:
    @pytest.fixture
    def _spec(self, request):
        # (cst, shape, backend, width) config
        raise NotImplementedError

    @pytest.fixture
    def spec(self, _spec):
        op = pxo.ConstantValued(shape=_spec[1], cst=_spec[0])
        ndi, width = _spec[2:]
        return op, ndi, width

    @pytest.fixture
    def cst(self, _spec) -> float:
        return _spec[0]

    @pytest.fixture
    def data_shape(self, _spec) -> pxt.OpShape:
        return _spec[1]

    @pytest.fixture
    def data_apply(self, data_shape, cst):
        codim, dim = data_shape
        return dict(
            in_=dict(arr=np.zeros(dim)),
            out=np.full(codim, fill_value=cst),
        )

    @pytest.fixture
    def data_grad(self, data_shape):
        _, dim = data_shape
        arr = conftest.MapT._random_array((dim,))
        return dict(
            in_=dict(arr=arr),
            out=np.zeros(dim),
        )

    @pytest.fixture
    def data_prox(self, data_shape):
        _, dim = data_shape
        arr = conftest.MapT._random_array((dim,))
        return dict(
            in_=dict(arr=arr, tau=2.1),
            out=arr,
        )

    @pytest.fixture
    def data_math_lipschitz(self, data_shape):
        _, dim = data_shape
        N_test = 5
        return conftest.MapT._random_array((N_test, dim))

    @pytest.fixture
    def data_math_diff_lipschitz(self, data_shape):
        _, dim = data_shape
        N_test = 5
        return conftest.MapT._random_array((N_test, dim))


class TestConstantValuedMap(ConstantValueMixin, conftest.DiffMapT):
    @pytest.fixture(
        params=itertools.product(
            [-3.14, 0, 2],  # cst
            [(2, 4), (3, 1)],  # shape
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param


class TestConstantValuedFunc(ConstantValueMixin, conftest.ProxDiffFuncT):
    disable_test = frozenset(
        conftest.ProxDiffFuncT.disable_test
        | {
            "test_interface_asloss",  # does not make sense for ConstantValued().
        }
    )

    @pytest.fixture(
        params=itertools.product(
            [-3.14, 0, 2],  # cst
            [(1, 4), (1, 1)],  # shape
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param



# How TransposeRule tests work:
#
# * TransposeRuleMixin auto-defines all arithmetic method (input,output) pairs.
#   [Caveat: we assume the base operators (op_orig) are correctly implemented.] (True if choosing test operators from examples/.)
#
# * To test a transposed-operator, inherit from TransposeRuleMixin and the suitable conftest.LinOpT
#   subclass which the transposed operator should abide by.


import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


# Test operators --------------------------------------------------------------
def op_linop(dim: int, codim_scale: int):
    import pyxu_tests.operator.examples.test_linop as tc

    return tc.Tile(N=dim, M=codim_scale)


def op_linfunc(dim: int):
    import pyxu_tests.operator.examples.test_linfunc as tc

    op = tc.ScaledSum(N=dim)
    return op


def op_squareop(dim: int):
    import pyxu_tests.operator.examples.test_squareop as tc

    return tc.CumSum(N=dim)


def op_normalop(dim: int):
    import pyxu_tests.operator.examples.test_normalop as tc

    rng = np.random.default_rng(seed=2)
    h = rng.normal(size=(dim,))
    return tc.CircularConvolution(h=h)


def op_unitop(dim: int):
    import pyxu_tests.operator.examples.test_unitop as tc

    return tc.Permutation(N=dim)


def op_selfadjointop(dim: int):
    import pyxu_tests.operator.examples.test_selfadjointop as tc

    return tc.SelfAdjointConvolution(N=dim)


def op_posdefop(dim: int):
    import pyxu_tests.operator.examples.test_posdefop as tc

    return tc.PSDConvolution(N=dim)


def op_projop(dim: int):
    import pyxu_tests.operator.examples.test_projop as tc

    return tc.Oblique(N=dim, alpha=np.pi / 4)


def op_orthprojop(dim: int):
    import pyxu_tests.operator.examples.test_orthprojop as tc

    return tc.ScaleDown(N=dim)


class TransposeRuleMixin:
    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        # Override in inherited class with the operator to be transposed.
        raise NotImplementedError

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, op_orig, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        op = op_orig.T
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, op_orig) -> pxt.OpShape:
        codim, dim = op_orig.shape
        return (dim, codim)

    @pytest.fixture
    def data_apply(self, op_orig) -> conftest.DataLike:
        x = self._random_array((op_orig.codim,))
        y = op_orig.adjoint(x)
        return dict(
            in_=dict(arr=x),
            out=y,
        )

    @pytest.fixture
    def data_adjoint(self, op_orig) -> conftest.DataLike:
        x = self._random_array((op_orig.dim,))
        y = op_orig.apply(x)
        return dict(
            in_=dict(arr=x),
            out=y,
        )

    @pytest.fixture
    def data_grad(self, op_orig) -> conftest.DataLike:
        # We know that linfuncs-to-be must have op.grad(x) = op.asarray()
        x = self._random_array((op_orig.codim,))
        y = op_orig.asarray().flatten()
        return dict(
            in_=dict(arr=x),
            out=y,
        )

    @pytest.fixture
    def data_prox(self, op_orig) -> conftest.DataLike:
        # We know that linfuncs-to-be must have op.prox(x, tau) = x - op.grad(x) * tau
        x = self._random_array((op_orig.codim,))
        g = op_orig.asarray().flatten()
        tau = np.abs(self._random_array((1,)))[0]
        y = x - tau * g
        return dict(
            in_=dict(
                arr=x,
                tau=tau,
            ),
            out=y,
        )


class TestTransposeRuleSquareOp(TransposeRuleMixin, conftest.SquareOpT):
    @pytest.fixture
    def op_orig(self):
        return op_squareop(7)


class TestTransposeRuleNormalOp(TransposeRuleMixin, conftest.NormalOpT):
    @pytest.fixture
    def op_orig(self):
        return op_normalop(7)


class TestTransposeRuleUnitOp(TransposeRuleMixin, conftest.UnitOpT):
    @pytest.fixture
    def op_orig(self):
        return op_unitop(7)


class TestTransposeRuleSelfAdjointOp(TransposeRuleMixin, conftest.SelfAdjointOpT):
    @pytest.fixture
    def op_orig(self):
        return op_selfadjointop(7)


class TestTransposeRulePosDefOp(TransposeRuleMixin, conftest.PosDefOpT):
    @pytest.fixture
    def op_orig(self):
        return op_posdefop(7)


class TestTransposeRuleProjOp(TransposeRuleMixin, conftest.ProjOpT):
    @pytest.fixture
    def op_orig(self):
        return op_projop(7)


class TestTransposeRuleOrthProjOp(TransposeRuleMixin, conftest.OrthProjOpT):
    @pytest.fixture
    def op_orig(self):
        return op_orthprojop(7)


class TestTransposeRuleLinOp(TransposeRuleMixin, conftest.LinOpT):
    @pytest.fixture(
        params=[
            op_linfunc(7),
            op_linop(2, 3),
        ]
    )
    def op_orig(self, request):
        return request.param

    @pytest.mark.parametrize("k", [1, 2])
    @pytest.mark.parametrize("which", ["LM", "SM"])
    def test_value1D_svdvals(self, op, ndi, _gpu, _op_svd, k, which):
        self._skip_if_disabled()
        if k > min(op.shape):
            pytest.skip(f"k={k} too high for {op.shape}-shaped operator.")
        super().test_value1D_svdvals(op, ndi, _gpu, _op_svd, k, which)


class TestTransposeRuleLinFunc(TransposeRuleMixin, conftest.LinFuncT):
    @pytest.fixture(
        params=[
            op_linop(1, 7),
            op_linfunc(1),
            op_squareop(1),
            op_normalop(1),
            op_unitop(1),
            op_selfadjointop(1),
            op_posdefop(1),
            # op_projop(1),  Not a projection op!
            # op_orthprojop(1),  Degenerate case; svd tests fail!
        ]
    )
    def op_orig(self, request):
        return request.param



import itertools

import numpy as np
import pytest
import scipy.ndimage as scimage
import skimage

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest
import pyxu_tests.operator.linop.diff.test_diff as test_diff


@pytest.mark.filterwarnings("ignore::numba.core.errors.NumbaPerformanceWarning")
class FilterMixin(conftest.SquareOpT):
    @pytest.fixture
    def data_shape(self, arg_shape) -> pxt.OpShape:
        dim = np.prod(arg_shape).item()
        return (dim, dim)

    @pytest.fixture
    def mode(self):
        return "constant"

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(
        self, arg_shape, mode, filter_klass, filter_kwargs, request
    ) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        with pxrt.Precision(width):
            op = filter_klass(
                arg_shape=arg_shape,
                mode=mode,
                gpu=ndi.name == "CUPY",
                dtype=width.value,
                **filter_kwargs,
            )
        return op, ndi, width


class TestMovingAverage(FilterMixin):
    @pytest.fixture
    def filter_klass(self):
        return pxo.MovingAverage

    @pytest.fixture(
        params=[
            # arg_shape, size, center, origin (for scipy)
            ((5,), 4, (0,), -2),
            ((5, 3, 4), 3, (0, 1, 2), (-1, 0, 1)),
            ((5, 3, 4), (5, 1, 3), None, 0),
        ]
    )
    def _spec(self, request):
        # (arg_shape, size, center, origin) configs to test.
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):
        arg_shape, _, _, _ = _spec
        return arg_shape

    @pytest.fixture
    def size(self, _spec):
        _, size, _, _ = _spec
        return size

    @pytest.fixture
    def center(self, _spec):
        _, _, center, _ = _spec
        return center

    @pytest.fixture
    def origin(self, _spec):
        _, _, _, origin = _spec
        return origin

    @pytest.fixture
    def filter_kwargs(self, size, center):
        return {"size": size, "center": center}

    @pytest.fixture
    def data_apply(self, _spec, mode) -> conftest.DataLike:
        arg_shape, size, _, origin = _spec
        arr = self._random_array(arg_shape)
        out = scimage.uniform_filter(arr, size=size, mode=mode, origin=origin)

        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestGaussian(FilterMixin):
    @pytest.fixture
    def filter_klass(self):
        return pxo.Gaussian

    @pytest.fixture(
        params=[
            # arg_shape, sigma, order, truncate
            ((8,), 3, 0, 1),
            ((4, 4, 4), 3, (0, 1, 2), 1),
            ((8, 8, 8), (1, 2, 3), 1, 2),
        ]
    )
    def _spec(self, request):
        # (arg_shape, sigma, order, truncate) configs to test.
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):
        arg_shape, _, _, _ = _spec
        return arg_shape

    @pytest.fixture
    def filter_kwargs(self, _spec):
        _, sigma, order, truncate = _spec
        return {"sigma": sigma, "truncate": truncate, "order": order}

    @pytest.fixture
    def data_apply(self, _spec, mode) -> conftest.DataLike:
        arg_shape, sigma, order, truncate = _spec
        arr = self._random_array(arg_shape)
        out = scimage.gaussian_filter(arr, sigma=sigma, truncate=truncate, order=order, mode=mode)
        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestDoG(FilterMixin):
    @pytest.fixture
    def filter_klass(self):
        return pxo.DifferenceOfGaussians

    @pytest.fixture(
        params=[
            # arg_shape, low_sigma, high_sigma, low_truncate, high_truncate
            ((8,), 2, 3, 1, 1),
            ((4, 4, 4), 1, (2, 1.5, 3), 1, 1),
            ((8, 8, 8), (1, 2, 3), 4, 1, 1),
        ]
    )
    def _spec(self, request):
        # (arg_shape, low_sigma, high_sigma, low_truncate, high_truncate) configs to test.
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):
        arg_shape, _, _, _, _ = _spec
        return arg_shape

    @pytest.fixture
    def filter_kwargs(self, _spec):
        _, low_sigma, high_sigma, low_truncate, high_truncate = _spec
        return {
            "low_sigma": low_sigma,
            "high_sigma": high_sigma,
            "low_truncate": low_truncate,
            "high_truncate": high_truncate,
        }

    @pytest.fixture
    def data_apply(self, _spec, mode) -> conftest.DataLike:
        arg_shape, low_sigma, high_sigma, low_truncate, high_truncate = _spec
        arr = self._random_array(arg_shape)
        out_low = scimage.gaussian_filter(arr, sigma=low_sigma, truncate=low_truncate, order=0, mode=mode)
        out_high = scimage.gaussian_filter(arr, sigma=high_sigma, truncate=high_truncate, order=0, mode=mode)
        out = out_low - out_high
        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestLaplace(FilterMixin):
    @pytest.fixture
    def filter_klass(self):
        return pxo.Laplace

    @pytest.fixture(params=[(8,), (4, 4, 4)])
    def arg_shape(self, request):
        return request.param

    @pytest.fixture
    def filter_kwargs(self):
        return dict()

    @pytest.fixture
    def data_apply(self, arg_shape, mode) -> conftest.DataLike:
        arr = self._random_array(arg_shape)
        out = scimage.laplace(arr, mode=mode)
        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


@pytest.mark.filterwarnings("ignore::numba.core.errors.NumbaPerformanceWarning")
class EdgeFilterMixin(conftest.DiffMapT):
    @pytest.fixture
    def data_shape(self, arg_shape) -> pxt.OpShape:
        dim = np.prod(arg_shape).item()
        return (dim, dim)

    @pytest.fixture
    def mode(self):
        return "constant"

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(
        self, arg_shape, mode, filter_klass, filter_kwargs, request
    ) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        with pxrt.Precision(width):
            op = filter_klass(
                arg_shape=arg_shape,
                mode=mode,
                gpu=ndi.name == "CUPY",
                dtype=width.value,
                **filter_kwargs,
            )
        return op, ndi, width

    @pytest.fixture(
        params=[
            # arg_shape, axis, axis_scipy
            ((4,), (0,), (0,)),
            ((4,), None, (0,)),
            ((4, 4, 4), (0, 2), (0, 2)),
            ((4, 4, 4), None, (0, 1, 2)),
        ]
    )
    def _spec(self, request):
        # arg_shape, axis, axis_scipy
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):
        arg_shape, _, _ = _spec
        return arg_shape

    @pytest.fixture
    def axis(self, _spec):
        _, axis, _ = _spec
        return axis

    @pytest.fixture
    def axis_skimage(self, _spec):
        _, _, axis_skimage = _spec
        return axis_skimage

    @pytest.fixture
    def filter_kwargs(self, axis):
        return {"axis": axis}

    @pytest.fixture
    def data_apply(self, arg_shape, filter_skimage, axis_skimage, mode) -> conftest.DataLike:
        arr = self._random_array(arg_shape)
        out = filter_skimage(arr, mode=mode, axis=axis_skimage)

        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestSobel(EdgeFilterMixin):
    @pytest.fixture
    def filter_klass(self):
        return pxo.Sobel

    @pytest.fixture
    def filter_skimage(self):
        return skimage.filters.sobel


class TestPrewitt(EdgeFilterMixin):
    @pytest.fixture
    def filter_klass(self):
        return pxo.Prewitt

    @pytest.fixture
    def filter_skimage(self):
        return skimage.filters.prewitt


class TestScharr(EdgeFilterMixin):
    @pytest.fixture
    def filter_klass(self):
        return pxo.Scharr

    @pytest.fixture
    def filter_skimage(self):
        return skimage.filters.scharr


class TestStructureTensor(conftest.DiffMapT):
    @pytest.fixture(
        params=[
            # 1,
            1.5,
            # 2
        ]
    )
    def sampling(self, request):
        return request.param

    @pytest.fixture(params=pxd.NDArrayInfo)
    def ndi(self, request):
        # [Sepand] Not inferred from spec(diff_kwargs) [unlike most Pyxu operators] since diff_kwargs() needs this information beforehand.
        ndi_ = request.param
        if ndi_.module() is None:
            pytest.skip(f"{ndi_} unsupported on this machine.")
        return ndi_

    @pytest.fixture(params=pxrt.Width)
    def width(self, request):
        # [Sepand] Not inferred from spec(diff_kwargs) [unlike most Pyxu operators] since diff_kwargs() needs this information beforehand.
        return request.param

    @pytest.fixture(params=["fd"])
    def diff_method(self, request):
        return request.param

    @pytest.fixture(
        params=[
            #  Finite Diff. ,   Gaussian Der.
            # (diff_typ, acc), (sigma, truncate)
            (("central", 2), (1.0, 1.0)),
        ]
    )
    def init_params(self, diff_method, sampling, request):
        params_fd, params_gd = request.param
        if diff_method == "fd":
            return test_diff.diff_params_fd(params_fd[0], params_fd[1], sampling)
        elif diff_method == "gd":
            return test_diff.diff_params_gd(params_gd[0], params_gd[1], sampling)

    @pytest.fixture
    def diff_params(self, init_params):
        return init_params[0]

    @pytest.fixture
    def gt_diffs(self, init_params):
        return init_params[1]

    @pytest.fixture
    def data_shape(self, arg_shape) -> pxt.OpShape:
        ndim = len(arg_shape)
        dim = np.prod(arg_shape).item()
        if ndim > 1:
            codim = (ndim * (ndim + 1) / 2) * dim
        else:
            codim = dim
        return (codim, dim)

    @pytest.fixture
    def mode(self):
        return "constant"

    @pytest.fixture
    def spec(self, arg_shape, filter_kwargs, ndi, width) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        with pxrt.Precision(width):
            op = pxo.StructureTensor(
                arg_shape=arg_shape,
                gpu=ndi == pxd.NDArrayInfo.CUPY,
                width=width,
                **filter_kwargs,
            )
        return op, ndi, width

    @pytest.fixture(params=[(8,), (4, 4), (4, 4, 4)])
    def arg_shape(self, request):
        return request.param

    @pytest.fixture
    def smooth_sigma(self):
        return 2.0

    @pytest.fixture
    def smooth_truncate(self):
        return 2.0

    @pytest.fixture(params=[False, True])
    def parallel(self, request):
        return request.param

    @pytest.fixture
    def filter_kwargs(self, diff_method, smooth_sigma, smooth_truncate, sampling, parallel, diff_params, mode):
        return {
            "diff_method": diff_method,
            "smooth_sigma": smooth_sigma,
            "smooth_truncate": smooth_truncate,
            "sampling": sampling,
            "parallel": parallel,
            "diff_params": diff_params,
            "mode": mode,
        }

    @pytest.fixture
    def data_apply(self, arg_shape, diff_method, gt_diffs, filter_kwargs) -> conftest.DataLike:
        arr = self._random_array(arg_shape)
        directions = [i for i in range(len(arg_shape))]
        # Cannot use skimage structure tensor as it uses the sobel filter as derivative.
        derivatives = test_diff.apply_gradient(
            arr,
            arg_shape=arg_shape,
            gt_diffs=gt_diffs,
            directions=directions,
            diff_method=diff_method,
            mode=filter_kwargs["mode"],
        )

        gt_smooth = test_diff.diff_params_gd(
            filter_kwargs["smooth_sigma"], filter_kwargs["smooth_truncate"], filter_kwargs["sampling"]
        )[1]

        out = []
        for der0, der1 in itertools.combinations_with_replacement(derivatives, 2):
            o = der0 * der1
            for dim in range(len(arg_shape)):
                o = test_diff.apply_derivative(
                    arr=o, arg_shape=arg_shape, axis=dim, gt_diffs=gt_smooth, order=0, mode=filter_kwargs["mode"]
                )
            out.append(o)
        out = np.concatenate(out)

        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )



import dask.array as da
import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.util as pxu


class TestGetArrayModule:
    def test_array(self, xp):
        x = xp.arange(5)
        assert pxu.get_array_module(x) is xp

    @pytest.mark.parametrize(
        ["obj", "fallback", "fail"],
        [
            [None, None, True],
            [None, np, False],
            [1, None, True],
            [1, np, False],
        ],
    )
    def test_fallback(self, obj, fallback, fail):
        # object is not an array type, so fail or return provided fallback
        if not fail:
            assert pxu.get_array_module(obj, fallback) is fallback
        else:
            with pytest.raises(ValueError):
                assert pxu.get_array_module(obj, fallback)


class TestRedirect:
    def function_function():
        def f(x, y):
            return "f"

        def g(x, y):
            return "g"

        f = pxu.redirect("x", NUMPY=g)(f)
        return f, g

    def function_staticmethod():
        def f(x, y):
            return "f"

        class Klass:
            @staticmethod
            def g(x, y):
                return "g"

        f = pxu.redirect("x", NUMPY=Klass.g)(f)
        return f, Klass.g

    def staticmethod_function():
        class Klass:
            @staticmethod
            def f(x, y):
                return "f"

        def g(x, y):
            return "g"

        Klass.f = pxu.redirect("x", NUMPY=g)(Klass.f)
        return Klass.f, g

    def staticmethod_staticmethod():
        class Klass:
            @staticmethod
            def f(x, y):
                return "f"

            @staticmethod
            def g(x, y):
                return "g"

        Klass.f = pxu.redirect("x", NUMPY=Klass.g)(Klass.f)
        return Klass.f, Klass.g

    def method_method():
        class Klass:
            def f(self, x, y):
                return "f"

            def g(self, x, y):
                return "g"

        klass = Klass()
        klass.f = pxu.redirect("x", NUMPY=klass.g)(klass.f)
        return klass.f, klass.g

    @pytest.fixture(
        params=[
            function_function(),  # function -> function
            function_staticmethod(),  # function -> staticmethod
            staticmethod_function(),  # staticmethod -> function
            staticmethod_staticmethod(),  # staticmethod -> staticmethod
            method_method(),  # method -> method
        ]
    )
    def callables(self, request):
        return request.param

    def test_invalid_signature(self, callables):
        f_decorated, f_target = callables
        with pytest.raises(ValueError):
            f_decorated(z=1)

    def test_nonArray_input(self, callables):
        f_decorated, f_target = callables
        with pytest.raises(ValueError):
            f_decorated(x=1, y=1)

    def test_dispatch_to_default(self, callables):
        f_decorated, f_target = callables

        kwargs = dict(x=da.array([1]), y=1)
        assert f_decorated(**kwargs) == "f"

    def test_dispatch_to_custom(self, callables):
        f_decorated, f_target = callables

        kwargs = dict(x=np.array([1]), y=1)
        assert f_decorated(**kwargs) == f_target(**kwargs)


class TestCompute:
    @pytest.fixture(
        params=[
            1,
            [1, 2, 3],
            np.arange(5),
            da.arange(5),
        ]
    )
    def single_input(self, request):
        return request.param

    def equal(self, x, y):
        if any(type(_) in pxd.supported_array_types() for _ in [x, y]):
            return np.allclose(x, y)
        else:
            return x == y

    @pytest.fixture(params=["compute", "persist"])
    def mode(self, request):
        return request.param

    def test_single_inputs(self, single_input, mode):
        cargs = pxu.compute(single_input, mode=mode)
        assert self.equal(cargs, single_input)

    def test_multi_inputs(self, mode):
        x = da.arange(5)
        y = x + 1
        i_args = (1, [1, 2, 3], np.arange(5), x, y)
        o_args = (1, [1, 2, 3], np.arange(5), np.arange(5), np.arange(1, 6))
        cargs = pxu.compute(*i_args, mode=mode)

        assert len(cargs) == len(o_args)
        for c, o in zip(cargs, o_args):
            assert self.equal(c, o)

    def test_invalid_mode(self):
        with pytest.raises(ValueError):
            pxu.compute(1, mode="test")

    def test_kwargs_does_not_fail(self):
        x = 1
        pxu.compute(x, optimize_graph=False)


class TestToNumPy:
    @pytest.fixture
    def arr(self) -> pxt.NDArray:
        return np.arange(-5, 5)

    @pytest.fixture
    def _arr(self, arr, xp):
        return xp.array(arr, dtype=arr.dtype)

    def test_backend_change(self, _arr):
        N = pxd.NDArrayInfo
        np_arr = pxu.to_NUMPY(_arr)

        assert N.from_obj(np_arr) == N.NUMPY
        if N.from_obj(_arr) == N.NUMPY:
            assert _arr is np_arr



import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class ReLU(pxa.Map):
    # f: \bR^{M} -> \bR^{M}
    #      x     -> max(x, 0)
    def __init__(self, M: int):
        super().__init__(shape=(M, M))
        self.lipschitz = np.inf

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr):
        y = arr.clip(min=0)
        return y


class TestReLU(conftest.MapT):
    @pytest.fixture(
        params=itertools.product(
            (  # dim, op
                (3, ReLU(M=3)),
                (1, ReLU(M=1)),
            ),
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (dim, dim)

    @pytest.fixture(
        params=[  # 2 test points
            dict(
                in_=dict(arr=np.array([-1, 0, 5])),
                out=np.array([0, 0, 5]),
            ),
            dict(
                in_=dict(arr=np.array([-3, 1, 5000])),
                out=np.array([0, 1, 5000]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test = 5
        return self._random_array((N_test, dim))



import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class CumSum(pxa.SquareOp):
    # f: \bR^{N} -> \bR^{N}
    #      x     -> [x1, x1+x2, ..., x1+...+xN]
    def __init__(self, N: int):
        super().__init__(shape=(N, N))
        self.lipschitz = np.sqrt(N * (N + 1) / 2)  # Frobenius norm

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr):
        y = arr.cumsum(axis=-1)
        return y

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr):
        y = arr[..., ::-1].cumsum(axis=-1)[..., ::-1]
        return y


class TestCumSum(conftest.SquareOpT):
    @pytest.fixture(
        params=itertools.product(
            ((10, CumSum(N=10)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (dim, dim)

    @pytest.fixture
    def data_apply(self, dim):
        arr = self._random_array((dim,))
        out = np.cumsum(arr)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )



import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.math as pxm
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class SquaredL2Norm(pxa.DiffFunc):
    # f: \bR^{M} -> \bR
    #      x     -> \norm{x}{2}^{2}
    def __init__(self, M: int):
        super().__init__(shape=(1, M))
        self.lipschitz = np.inf
        self.diff_lipschitz = 2

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr):
        y = pxm.norm(arr, axis=-1, keepdims=True)
        y **= 2
        return y

    @pxrt.enforce_precision(i="arr")
    def grad(self, arr):
        return 2 * arr

    def asloss(self, data: pxt.NDArray = None) -> pxt.OpT:
        from pyxu.operator import shift_loss

        op = shift_loss(op=self, data=data)
        return op


class TestSquaredL2Norm(conftest.DiffFuncT):
    @pytest.fixture(
        params=itertools.product(
            ((4, SquaredL2Norm(M=4)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((4,))),
                out=np.zeros((1,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 1)),
                out=np.array([14]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test, dim = 6, self._sanitize(dim, 3)
        return self._random_array((N_test, dim), seed=5)

    @pytest.fixture
    def data_math_diff_lipschitz(self, dim):
        N_test, dim = 6, self._sanitize(dim, 3)
        return self._random_array((N_test, dim), seed=6)

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((4,))),
                out=np.zeros((4,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 1)),
                out=2 * np.arange(-3, 1),
            ),
        ]
    )
    def data_grad(self, request):
        return request.param



from pyxu.info.plugin import _load_entry_points
from pyxu.operator.blocks import *
from pyxu.operator.func.indicator import *
from pyxu.operator.func.loss import *
from pyxu.operator.func.norm import *
from pyxu.operator.linop import *
from pyxu.operator.linop.base import *
from pyxu.operator.linop.diff import *
from pyxu.operator.linop.fft.fft import *
from pyxu.operator.linop.fft.nufft import *
from pyxu.operator.linop.filter import *
from pyxu.operator.linop.kron import *
from pyxu.operator.linop.pad import *
from pyxu.operator.linop.reduce import *
from pyxu.operator.linop.select import *
from pyxu.operator.linop.stencil._stencil import _Stencil
from pyxu.operator.linop.stencil.stencil import *
from pyxu.operator.map.base import *
from pyxu.operator.map.ufunc import *

_load_entry_points(globals(), group="pyxu.operator")



import inspect

import numpy as np
import pytest

import pyxu.util as pxu


class TestVectorize:
    @pytest.fixture(
        params=[
            lambda x: (x.sum(keepdims=True) + 1).astype(np.half),  # 1D only
            lambda x, y: (x.sum(keepdims=True) + y).astype(np.half),  # 1D only, multi-parameter
            lambda x, y=1: (x.sum(keepdims=True) + y).astype(np.half),  # 1D only, multi-parameter (with defaults)
            lambda x: (x.sum(axis=-1, keepdims=True) + 1).astype(np.half),  # already has desired ND behaviour
        ]
    )
    def func(self, request):
        return request.param

    @pytest.fixture(
        params=[  # (method, codim)  [codim-size when specified comes from `data_func` fixture.]
            ("scan", None),
            ("scan", 10),  # codim irrelevant; bogus value to see if unused
            ("parallel", 5),
            ("scan_dask", 5),
        ]
    )
    def vfunc(self, func, request):
        method, codim = request.param
        decorate = pxu.vectorize(i="x", method=method, codim=codim)
        return decorate(func)

    @pytest.fixture
    def data_func(self, func):
        sig = inspect.Signature.from_callable(func)
        data = dict(
            in_=dict(x=np.arange(5)),
            out=np.array([11]),
        )
        if "y" in sig.parameters:
            data["in_"].update(y=1)
        return data

    def test_1d(self, vfunc, data_func):
        out_gt = data_func["out"]

        in_ = data_func["in_"]
        out = vfunc(**in_)

        assert out.ndim == 1
        assert np.allclose(out, out_gt)

    def test_nd(self, vfunc, data_func):
        sh_extra = (2, 1)  # prepend input/output shape by this amount.

        out_gt = data_func["out"]
        out_gt = np.broadcast_to(out_gt, (*sh_extra, out_gt.shape[-1]))

        in_ = data_func["in_"]
        in_["x"] = np.broadcast_to(in_["x"], (*sh_extra, *in_["x"].shape))
        out = vfunc(**in_)

        assert out.ndim == out_gt.ndim
        assert np.allclose(out, out_gt)

    def test_precision(self, func, vfunc, data_func):
        # decorated function should have same output dtype as base function.
        in_ = data_func["in_"]

        out_f = func(**in_)
        out_vf = vfunc(**in_)

        assert out_f.dtype == out_vf.dtype

    def test_backend(self, vfunc, data_func, xp):
        # returned array from decorated function should have same type as input array.
        in_ = data_func["in_"]
        in_["x"] = xp.array(in_["x"])
        out = vfunc(**in_)

        assert type(out) == type(in_["x"])  # noqa: E721



import typing as typ

import numpy as np

import pyxu.abc.operator as pxo
import pyxu.info.ptype as pxt
import pyxu.util as pxu


class SquaredL2Norm(pxo.DiffFunc):
    def __init__(self, shape: pxt.OpShape = None):
        super(SquaredL2Norm, self).__init__(shape=(1, None))
        self._diff_lipschitz = 2

    def apply(self, arr: pxt.NDArray) -> pxt.Real:
        xp = pxu.get_array_module(arr)
        return xp.linalg.norm(arr, axis=-1, keepdims=True) ** 2

    def grad(self, arr: pxt.NDArray) -> pxt.NDArray:
        return 2 * arr


class L1Norm(pxo.ProxFunc):
    def __init__(self, shape=None):
        super(L1Norm, self).__init__(shape=(1, None))
        self._lipschitz = 1

    def apply(self, arr: pxt.NDArray) -> pxt.Real:
        xp = pxu.get_array_module(arr)
        return xp.linalg.norm(arr, ord=1, axis=-1, keepdims=True)

    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return (abs(arr) - tau).clip(0, None) * xp.sign(arr)


class FirstDerivative(pxo.LinOp):
    def __init__(self, size: int, axis: int = -1, sampling: float = 1.0, edge: bool = True, kind: str = "forward"):
        super(FirstDerivative, self).__init__((size, size))
        self.axis = axis
        self.sampling = sampling
        self.edge = edge
        self.kind = kind

        # choose apply and adjoint kind
        if kind == "forward":
            self.apply = self._apply_forward
            self.adjoint = self._adjoint_forward
        elif kind == "centered":
            self.apply = self._apply_centered
            self.adjoint = self._adjoint_centered
        elif kind == "backward":
            self.apply = self._apply_backward
            self.adjoint = self._adjoint_backward
        else:
            raise NotImplementedError("kind must be forward, centered, " "or backward")

    def _apply_forward(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        arr = xp.swapaxes(arr, self.axis, -1)
        y = xp.zeros_like(arr)
        y[..., :-1] = (arr[..., 1:] - arr[..., :-1]) / self.sampling
        return y

    def _adjoint_forward(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        arr = xp.swapaxes(arr, self.axis, -1)
        y = xp.zeros_like(arr)
        y[..., :-1] -= arr[..., :-1] / self.sampling
        y[..., 1:] += arr[..., :-1] / self.sampling
        return y

    def _apply_centered(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        arr = xp.swapaxes(arr, self.axis, -1)
        y = xp.zeros_like(arr)
        y[..., 1:-1] = 0.5 * (arr[..., 2:] - arr[..., :-2]) / self.sampling
        if self.edge:
            y[..., 0] = (arr[..., 1] - arr[..., 0]) / self.sampling
            y[..., 0] = (arr[..., -1] - arr[..., -2]) / self.sampling
        return y

    def _adjoint_centered(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        arr = xp.swapaxes(arr, self.axis, -1)
        y = xp.zeros_like(arr)
        y[..., :-2] -= 0.5 * arr[..., 1:-1] / self.sampling
        y[..., 2:] += 0.5 * arr[..., 1:-1] / self.sampling
        if self.edge:
            y[..., 0] -= arr[..., 0] / self.sampling
            y[..., 1] += arr[..., 0] / self.sampling
            y[..., -2] -= arr[..., -1] / self.sampling
            y[..., -1] += arr[..., -1] / self.sampling
        return y

    def _apply_backward(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        arr = xp.swapaxes(arr, self.axis, -1)
        y = xp.zeros_like(arr)
        y[..., 1:] = (arr[..., 1:] - arr[..., :-1]) / self.sampling
        return y

    def _adjoint_backward(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        arr = xp.swapaxes(arr, self.axis, -1)
        y = xp.zeros_like(arr)
        y[..., :-1] -= arr[..., 1:] / self.sampling
        y[..., 1:] += arr[..., 1:] / self.sampling
        return y


class Masking(pxo.LinOp):
    def __init__(self, size: int, sampling_bool: typ.Union[pxt.NDArray, list]):
        if isinstance(sampling_bool, list):
            import numpy as xp
        else:
            xp = pxu.get_array_module(sampling_bool)

        self.sampling_bool = xp.asarray(sampling_bool).reshape(-1).astype(bool)
        self.input_size = size
        self.nb_of_samples = self.sampling_bool[self.sampling_bool == True].size
        if self.sampling_bool.size != size:
            raise ValueError("Invalid size of boolean sampling array.")
        super(Masking, self).__init__(shape=(self.nb_of_samples, self.input_size))

    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        return arr[..., self.sampling_bool]

    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        arr = arr.reshape(1, -1) if (arr.ndim == 1) else arr
        y = xp.zeros((*arr.shape[:-1], self.input_size), dtype=arr.dtype)
        y[..., self.sampling_bool] = arr
        return y


class DownSampling(Masking):
    def __init__(
        self,
        size: int,
        downsampling_factor: typ.Union[int, tuple, list],
        shape: typ.Optional[tuple] = None,
        axis: typ.Optional[int] = None,
    ):
        if type(downsampling_factor) is int:
            if (shape is not None) and (axis is None):
                self.downsampling_factor = len(shape) * (downsampling_factor,)
            else:
                self.downsampling_factor = (downsampling_factor,)
        else:
            self.downsampling_factor = tuple(downsampling_factor)
        if shape is not None:
            if size != np.prod(shape):
                raise ValueError(f"Array size {size} is incompatible with array shape {shape}.")
            if (axis is not None) and (axis > len(shape) - 1):
                raise ValueError(f"Array size {size} is incompatible with array shape {shape}.")
        if (shape is None) and (len(self.downsampling_factor) > 1):
            raise ValueError("Please specify an array shape for multi-dimensional downsampling.")
        elif (shape is not None) and (axis is None) and (len(shape) != len(self.downsampling_factor)):
            raise ValueError(f"Inconsistent downsampling factors {downsampling_factor} for array of shape {shape}.")
        self.input_size = size
        self.input_shape = shape
        self.axis = axis
        self.downsampling_mask = self.compute_downsampling_mask()
        if self.input_shape is None:
            self.output_shape = None
        else:
            if len(self.downsampling_factor) > 1:
                output_shape = []
                for ax in range(len(self.input_shape)):
                    axis_indices = np.arange(self.input_shape[ax])
                    downsampled_axis_indices = axis_indices % self.downsampling_factor[ax]
                    output_shape.append(downsampled_axis_indices[downsampled_axis_indices == 0].size)
                self.output_shape = tuple(output_shape)
            else:
                output_shape = list(self.input_shape)
                downsampled_axis_indices = np.arange(self.input_shape[self.axis])
                downsampled_axis_indices = downsampled_axis_indices % self.downsampling_factor
                output_shape[self.axis] = downsampled_axis_indices[downsampled_axis_indices == 0].size
                self.output_shape = tuple(output_shape)

        super(DownSampling, self).__init__(size=self.input_size, sampling_bool=self.downsampling_mask)

    def compute_downsampling_mask(self) -> np.ndarray:
        if self.input_shape is None:
            indices = np.arange(self.input_size)
            downsampled_mask = (indices % self.downsampling_factor) == 0
        else:
            if len(self.downsampling_factor) > 1:
                downsampled_mask = True
                for ax in range(len(self.input_shape)):
                    axis_indices = np.arange(self.input_shape[ax])
                    downsampled_axis_indices = axis_indices % self.downsampling_factor[ax]
                    downsampled_axis_indices = downsampled_axis_indices.reshape(
                        downsampled_axis_indices.shape + (len(self.input_shape) - 1) * (1,)
                    )
                    downsampled_axis_indices = np.swapaxes(downsampled_axis_indices, 0, ax)
                    downsampled_mask = downsampled_mask * (downsampled_axis_indices == 0)
            else:
                downsampled_mask = np.zeros(shape=self.input_shape, dtype=bool)
                downsampled_mask = np.swapaxes(downsampled_mask, 0, self.axis)
                downsampled_axis_indices = np.arange(self.input_shape[self.axis])
                downsampled_axis_indices = downsampled_axis_indices % self.downsampling_factor
                downsampled_mask[downsampled_axis_indices == 0, ...] = True
                downsampled_mask = np.swapaxes(downsampled_mask, 0, self.axis)
        return downsampled_mask.reshape(-1)



import inspect

import numpy as np

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu.util as pxu
from pyxu.util.operator import _dask_zip

__all__ = [
    "FFT",
]


class FFT(pxa.LinOp):  # Inherits from LinOp instead of NormalOp since operator not square if `real=True`.
    r"""
    Multi-dimensional Discrete Fourier Transform (DFT) :math:`A: \mathbb{C}^{N_{1} \times \cdots \times N_{D}} \to
    \mathbb{C}^{N_{1} \times \cdots \times N_{D}}`.

    The FFT is defined as follows:

    .. math::

       (A \, \mathbf{x})[\mathbf{k}]
       =
       \sum_{\mathbf{n}} \mathbf{x}[\mathbf{n}]
       \exp\left[-j 2 \pi \langle \frac{\mathbf{n}}{\mathbf{N}}, \mathbf{k} \rangle \right],

    .. math::

       (A^{*} \, \hat{\mathbf{x}})[\mathbf{n}]
       =
       \sum_{\mathbf{k}} \hat{\mathbf{x}}[\mathbf{k}]
       \exp\left[j 2 \pi \langle \frac{\mathbf{n}}{\mathbf{N}}, \mathbf{k} \rangle \right],

    .. math::

       (\mathbf{x}, \, \hat{\mathbf{x}}) \in \mathbb{C}^{N_{1} \times \cdots \times N_{D}},
       \quad
       (\mathbf{n}, \, \mathbf{k}) \in \{0, \ldots, N_{1}-1\} \times \cdots \times \{0, \ldots, N_{D}-1\}.

    The DFT is taken over any number of axes by means of the Fast Fourier Transform algorithm (FFT).


    .. rubric:: Implementation Notes

    * The CPU implementation uses `SciPy's FFT implementation <https://docs.scipy.org/doc/scipy/reference/fft.html>`_.
    * The GPU implementation uses cuFFT via `CuPy <https://docs.cupy.dev/en/latest/reference/scipy_fft.html>`_.


    Examples
    --------

    * 1D DFT of a cosine pulse.

      .. code-block:: python3

         from pyxu.operator import FFT
         import pyxu.util as pxu

         N = 10
         op = FFT(
             arg_shape=(N,),
             real=True,
         )

         x = np.cos(2 * np.pi / N * np.arange(N))
         y = pxu.view_as_complex(op.apply(x))
         # [0, N/2, 0, 0, 0, 0, 0, 0, 0, N/2]

         z = op.adjoint(op.apply(x))
         # np.allclose(z, N * x) -> True

    * 1D DFT of a complex exponential pulse.

      .. code-block:: python3

         from pyxu.operator import FFT
         import pyxu.util as pxu

         N = 10
         op = FFT(
             arg_shape=(N,),
             real=False,  # complex-valued inputs
         )

         x = np.exp(1j * 2 * np.pi / N * np.arange(N))
         y = pxu.view_as_complex(
                 op.apply(
                     pxu.view_as_real(x)
                 )
             )
         # [0, N, 0, 0, 0, 0, 0, 0, 0, 0]

         z = pxu.view_as_complex(
                 op.adjoint(
                     op.apply(
                         pxu.view_as_real(x)
                     )
                 )
             )
         # np.allclose(z, N * x) -> True

    * 2D DFT of an image

      .. code-block:: python3

         from pyxu.operator import FFT
         import pyxu.util as pxu

         N_h, N_w = 10, 8
         op = FFT(
             arg_shape=(N_h, N_w),
             real=True,
         )

         x = np.pad(  # an image
             np.ones((N_h//2, N_w//2)),
             pad_width=((0, N_h//2), (0, N_w//2)),
         )
         y = pxu.view_as_complex(  # sinc
                 op.apply(x.reshape(-1))
             ).reshape(N_h, N_w)
         z = op.adjoint(
                 op.apply(x.reshape(-1))
             ).reshape(N_h, N_w)
         # np.allclose(z, (N_h * N_w) * x) -> True

    * 1D DFT of an image's rows

      .. code-block:: python3

         from pyxu.operator import FFT
         import pyxu.util as pxu

         N_h, N_w = 10, 8
         op = FFT(
             arg_shape=(N_h, N_w),
             axes=-1,
             real=True,
         )

         x = np.pad(  # an image
             np.ones((N_h//2, N_w//2)),
             pad_width=((0, N_h//2), (0, N_w//2)),
         )
         y = pxu.view_as_complex(  # sinc
                 op.apply(x.reshape(-1))
             ).reshape(N_h, N_w)
         z = op.adjoint(
                 op.apply(x.reshape(-1))
             ).reshape(N_h, N_w)
         # np.allclose(z, N_w * x) -> True
    """

    def __init__(
        self,
        arg_shape: pxt.NDArrayShape,
        axes: pxt.NDArrayAxis = None,
        real: bool = False,
        **kwargs,
    ):
        r"""
        Parameters
        ----------
        arg_shape: NDArrayShape
            (N_1, ..., N_D) shape of the input array :math:`\mathbf{x} \in \mathbb{R}^{N_{1} \times \cdots \times
            N_{D}}` or :math:`\mathbb{C}^{N_{1} \times \cdots \times N_{D}}`.
        axes: NDArrayAxis
            Axis or axes along which the DFT is performed.  The default, axes=None, will transform all dimensions of the
            input array.
        real: bool
            If ``True``, assumes :py:func:`~pyxu.operator.FFT.apply` takes (..., N.prod()) inputs in
            :math:`\mathbb{R}^{N}`.

            If ``False``, then :py:func:`~pyxu.operator.FFT.apply` takes (..., 2N.prod()) inputs, i.e.
            :math:`\mathbb{C}^{N}` vectors viewed as bijections with :math:`\mathbb{R}^{2N}`.
        kwargs: dict
            Extra kwargs passed to :py:func:`scipy.fft.fftn` or :py:func:`cupyx.scipy.fft.fftn`.

            Supported parameters for :py:func:`scipy.fft.fftn` are:

                * workers: int = 1

            Supported parameters for :py:func:`cupyx.scipy.fft.fftn` are:

                * NOT SUPPORTED FOR NOW

            Default values are chosen if unspecified.
        """
        arg_shape = pxu.as_canonical_shape(arg_shape)
        N_dim, N = len(arg_shape), np.prod(arg_shape)

        if axes is None:
            axes = tuple(range(N_dim))
        axes = np.unique(pxu.as_canonical_shape(axes))  # drop potential duplicates
        assert np.all((-N_dim <= axes) & (axes < N_dim))  # all axes in valid range
        axes = (axes + N_dim) % N_dim  # get rid of negative axes

        sh_op = [2 * N, 2 * N]
        sh_op[1] //= 2 if real else 1
        super().__init__(shape=sh_op)

        self._arg_shape = tuple(arg_shape)
        self._axes = tuple(axes)
        self._real = bool(real)
        self._kwargs = {
            pxd.NDArrayInfo.NUMPY: dict(
                workers=kwargs.get("workers", 1),
            ),
            pxd.NDArrayInfo.CUPY: dict(),
        }

        self.lipschitz = self.estimate_lipschitz()

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        arg_shape = np.array(self._arg_shape, dtype=int)
        axes = np.array(self._axes, dtype=int)

        L = np.sqrt(arg_shape[axes].prod())
        return L

    def gram(self) -> pxt.OpT:
        from pyxu.operator import HomothetyOp

        op_g = HomothetyOp(dim=self.dim, cst=self.lipschitz**2)
        return op_g

    def cogram(self) -> pxt.OpT:
        if self._real:
            # Enforcing real-valued adjoint() outputs implies a projection between A and A.T.
            # There is no simple closed form to compute the co-gram in this case.
            op_cg = super().cogram()
        else:
            from pyxu.operator import HomothetyOp

            op_cg = HomothetyOp(dim=self.codim, cst=self.lipschitz**2)
        return op_cg

    @pxrt.enforce_precision(i=("arr", "damp"))
    def pinv(self, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
        N = self.lipschitz**2
        out = self.adjoint(arr)
        out /= N + damp
        return out

    def dagger(self, damp: pxt.Real, **kwargs) -> pxt.OpT:
        N = self.lipschitz**2
        op = self.T / (N + damp)
        return op

    @pxrt.enforce_precision()
    def svdvals(self, **kwargs) -> pxt.NDArray:
        D = pxa.UnitOp.svdvals(self, **kwargs) * self.lipschitz
        return D

    def _transform(self, arr: pxt.NDArray, mode: str) -> pxt.NDArray:
        # Internal method for apply/adjoint.
        #
        # Parameters
        # ----------
        # arr: pxt.NDArray [float/complex]
        #     (N_1, ..., N_D) array to transform.
        # mode: str
        #     Transform direction:
        #
        #     * 'fw': fftn(, norm="backward")
        #     * 'bw': ifftn(, norm="forward")
        #
        # Returns
        # -------
        # out: pxt.NDArray [complex]
        #     (N_1, ..., N_D) transformed array.
        N = pxd.NDArrayInfo  # shorthand
        ndi = N.from_obj(arr)

        if ndi == N.NUMPY:
            fft = pxu.import_module("scipy.fft")
        elif ndi == N.CUPY:
            fft = pxu.import_module("cupyx.scipy.fft")
        else:
            raise ValueError("Unknown NDArray category.")

        func, norm = dict(  # ref: scipy.fft norm conventions
            fw=(fft.fftn, "backward"),
            bw=(fft.ifftn, "forward"),
        )[mode]

        # `self._kwargs()` contains parameters undersood by different FFT backends.
        # If we must fallback to `scipy.fft`, need to drop all non-standard parameters.
        sig = inspect.Signature.from_callable(func)
        kwargs = {k: v for (k, v) in self._kwargs[ndi].items() if (k in sig.parameters)}

        out = func(
            x=arr,
            axes=self._axes,
            norm=norm,
            **kwargs,
        )
        return out

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        r"""
        Parameters
        ----------
        arr: NDArray
            * (...,  N.prod()) inputs :math:`\mathbf{x} \in \mathbb{R}^{N_{1} \times \cdots \times N_{D}}`
              (``real=True``.)
            * (..., 2N.prod()) inputs :math:`\mathbf{x} \in \mathbb{C}^{N_{1} \times \cdots \times N_{D}}` viewed as a
              real array. (See :py:func:`~pyxu.util.view_as_real`.)

        Returns
        -------
        out: NDArray
            (..., 2N.prod()) outputs :math:`\hat{\mathbf{x}} \in \mathbb{C}^{N_{1} \times \cdots \times N_{D}}` viewed
            as a real array. (See :py:func:`~pyxu.util.view_as_real`.)
        """
        if self._real:
            r_width = pxrt.Width(arr.dtype)
            c_dtype = r_width.complex.value
        else:
            arr = pxu.view_as_complex(arr)
            c_dtype = arr.dtype

        sh = arr.shape[:-1]
        N_trans = np.prod(sh, dtype=int)
        func = lambda _: self._transform(arr=_, mode="fw")
        blks = _dask_zip(
            func=(func,) * N_trans,
            data=arr.reshape(N_trans, *self._arg_shape),
            out_shape=(self._arg_shape,) * N_trans,
            out_dtype=(c_dtype,) * N_trans,
            parallel=True,
        )

        xp = pxu.get_array_module(arr)
        out = xp.stack(blks, axis=0).reshape(*sh, -1)

        return pxu.view_as_real(out)

    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        r"""
        Parameters
        ----------
        arr: NDArray
            (..., 2N.prod()) outputs :math:`\hat{\mathbf{x}} \in \mathbb{C}^{N_{1} \times \cdots \times N_{D}}` viewed
            as a real array. (See :py:func:`~pyxu.util.view_as_real`.)

        Returns
        -------
        out: NDArray
            * (...,  N.prod()) inputs :math:`\mathbf{x} \in \mathbb{R}^{N_{1} \times \cdots \times N_{D}}`
              (``real=True``.)
            * (..., 2N.prod()) inputs :math:`\mathbf{x} \in \mathbb{C}^{N_{1} \times \cdots \times N_{D}}` viewed as a
              real array. (See :py:func:`~pyxu.util.view_as_real`.)
        """
        arr = pxu.view_as_complex(arr)
        c_dtype = arr.dtype

        sh = arr.shape[:-1]
        N_trans = np.prod(sh, dtype=int)
        func = lambda _: self._transform(arr=_, mode="bw")
        blks = _dask_zip(
            func=(func,) * N_trans,
            data=arr.reshape(N_trans, *self._arg_shape),
            out_shape=(self._arg_shape,) * N_trans,
            out_dtype=(c_dtype,) * N_trans,
            parallel=True,
        )

        xp = pxu.get_array_module(arr)
        out = xp.stack(blks, axis=0).reshape(*sh, -1)

        if self._real:
            return out.real
        else:
            return pxu.view_as_real(out)



import collections.abc as cabc
import datetime as dt
import typing as typ
import warnings

import numpy as np

import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.util as pxu
from pyxu.info.plugin import _load_entry_points

__all__ = [
    "AbsError",
    "ManualStop",
    "MaxDuration",
    "MaxIter",
    "MaxCarbon",
    "Memorize",
    "RelError",
]

__all__ = _load_entry_points(globals(), group="pyxu.opt.stop", names=__all__)

SVFunction = typ.Union[
    cabc.Callable[[pxt.Real], pxt.Real],
    cabc.Callable[[pxt.NDArray], pxt.NDArray],
]


class MaxIter(pxa.StoppingCriterion):
    """
    Stop iterative solver after a fixed number of iterations.

    .. note::

        If you want to add a grace period to a solver, i.e. for it to do *at least* N iterations before stopping based
        on the value of another criteria, you can AND :py:class:`~pyxu.opt.stop.MaxIter` with the other criteria.

        .. code-block:: python3

           sc = MaxIter(n=5) & AbsError(eps=0.1)
           # If N_iter < 5  -> never stop.
           # If N_iter >= 5 -> stop if AbsError() decides to.
    """

    def __init__(self, n: pxt.Integer):
        """
        Parameters
        ----------
        n: Integer
            Max number of iterations allowed.
        """
        try:
            assert int(n) > 0
            self._n = int(n)
        except Exception:
            raise ValueError(f"n: expected positive integer, got {n}.")
        self._i = 0

    def stop(self, state: cabc.Mapping) -> bool:
        self._i += 1
        return self._i > self._n

    def info(self) -> cabc.Mapping[str, float]:
        return dict(N_iter=self._i)

    def clear(self):
        self._i = 0


class ManualStop(pxa.StoppingCriterion):
    """
    Continue-forever criterion.

    This class is useful when calling :py:meth:`~pyxu.abc.Solver.fit` with mode=MANUAL/ASYNC to defer the stopping
    decision to an explicit call by the user, i.e.:

    * mode=MANUAL: user must stop calling ``next(solver.steps())``;
    * mode=ASYNC: user must call :py:meth:`~pyxu.abc.Solver.stop`.
    """

    def stop(self, state: cabc.Mapping) -> bool:
        return False

    def info(self) -> cabc.Mapping[str, float]:
        return dict()

    def clear(self):
        pass


class MaxDuration(pxa.StoppingCriterion):
    """
    Stop iterative solver after a specified duration has elapsed.
    """

    def __init__(self, t: dt.timedelta):
        """
        Parameters
        ----------
        t: ~datetime.timedelta
            Max runtime allowed.
        """
        try:
            assert t > dt.timedelta()
            self._t_max = t
        except Exception:
            raise ValueError(f"t: expected positive duration, got {t}.")
        self._t_start = dt.datetime.now()
        self._t_now = self._t_start

    def stop(self, state: cabc.Mapping) -> bool:
        self._t_now = dt.datetime.now()
        return (self._t_now - self._t_start) > self._t_max

    def info(self) -> cabc.Mapping[str, float]:
        d = (self._t_now - self._t_start).total_seconds()
        return dict(duration=d)

    def clear(self):
        self._t_start = dt.datetime.now()
        self._t_now = self._t_start


class MaxCarbon(pxa.StoppingCriterion):
    """
    Stop iterative solver after a threshold amount of carbon dioxide (CO2) was produced by computing resources used to
    execute the optimization.

    .. note::

       Codecarbon is not a core dependency of Pyxu and must be installed separately.

       On Windows and MacOS, `codecarbon <https://pypi.org/project/codecarbon/>`_ tracks power consumption of Intel
       processors using the `Intel Power Gadget (IPG)
       <https://www.intel.com/content/www/us/en/developer/articles/tool/power-gadget.html>`_.  IPG must be installed
       independently.  (MacOS-specific: IPG must also have correct security permissions.)
    """

    def __init__(self, co2: pxt.Real):
        """
        Parameters
        ----------
        co2: Real
            Max allowed CO2 emissions [Kg].
        """
        try:
            assert float(co2) > 0
            self._co2_max = float(co2)
        except Exception:
            raise ValueError(f"co2: expected positive carbon quantity, got {co2}.")

        codecarbon = pxu.import_module("codecarbon")
        self.tracker = codecarbon.EmissionsTracker(
            api_call_interval=-1,
            save_to_file=False,
            log_level="warning",
        )
        self.tracker.start()
        self._co2_start = self.tracker.flush()
        self._co2_now = self._co2_start

    def stop(self, state: cabc.Mapping) -> bool:
        self._co2_now = self.tracker.flush()
        decision = (self._co2_now - self._co2_start) > self._co2_max
        if decision:
            self.tracker.stop()
        return decision

    def info(self) -> cabc.Mapping[str, float]:
        return dict(co2=self._co2_now)

    def clear(self):
        self.tracker.start()
        self._co2_start = self.tracker.flush()
        self._co2_now = self._co2_start


class Memorize(pxa.StoppingCriterion):
    """
    Memorize a variable.  (Special :py:class:`~pyxu.abc.StoppingCriterion` mostly useful for tracking objective
    functions in :py:class:`~pyxu.abc.Solver`.)
    """

    def __init__(self, var: pxt.VarName):
        """
        Parameters
        ----------
        var: VarName
            Variable in :py:attr:`pyxu.abc.Solver._mstate` to query.  Must be a scalar or NDArray (1D).
        """
        self._var = var
        self._val = np.r_[0]  # last memorized value in stop().

    def stop(self, state: cabc.Mapping) -> bool:
        x = state[self._var]
        if isinstance(x, pxt.Real):
            x = np.r_[x]
        assert x.ndim == 1

        self._val = pxu.compute(x)
        return False

    def info(self) -> cabc.Mapping[str, float]:
        if self._val.size == 1:
            data = {f"Memorize[{self._var}]": float(self._val.max())}  # takes the only element available.
        else:
            data = {
                f"Memorize[{self._var}]_min": float(self._val.min()),
                f"Memorize[{self._var}]_max": float(self._val.max()),
            }
        return data

    def clear(self):
        self._val = np.r_[0]


class AbsError(pxa.StoppingCriterion):
    """
    Stop iterative solver after absolute norm of a variable (or function thereof) reaches threshold.
    """

    def __init__(
        self,
        eps: pxt.Real,
        var: pxt.VarName = "x",
        f: SVFunction = None,
        norm: pxt.Real = 2,
        satisfy_all: bool = True,
    ):
        """
        Parameters
        ----------
        eps: Real
            Positive threshold.
        var: VarName
            Variable in :py:attr:`pyxu.abc.Solver._mstate` to query.
        f: ~collections.abc.Callable
            Optional function to pre-apply to ``_mstate[var]`` before applying the norm.  Defaults to the identity
            function. The callable should either:

            * accept a scalar input -> output a scalar, or
            * accept an NDArray input -> output an NDArray, i.e same semantics as :py:meth:`~pyxu.abc.Map.apply`.
        norm: Integer, Real
            Ln norm to use >= 0. (Default: L2.)
        satisfy_all: bool
            If True (default) and ``_mstate[var]`` is multi-dimensional, stop if all evaluation points lie below
            threshold.
        """
        try:
            assert eps > 0
            self._eps = eps
        except Exception:
            raise ValueError(f"eps: expected positive threshold, got {eps}.")

        self._var = var
        self._f = f if (f is not None) else (lambda _: _)

        try:
            assert norm >= 0
            self._norm = norm
        except Exception:
            raise ValueError(f"norm: expected non-negative, got {norm}.")

        self._satisfy_all = satisfy_all
        self._val = np.r_[0]  # last computed Ln norm(s) in stop().

    def stop(self, state: cabc.Mapping) -> bool:
        fx = self._f(state[self._var])
        if isinstance(fx, pxt.Real):
            fx = np.r_[fx]
        xp = pxu.get_array_module(fx)

        self._val = xp.linalg.norm(fx, ord=self._norm, axis=-1, keepdims=True)
        rule = xp.all if self._satisfy_all else xp.any
        decision = rule(self._val <= self._eps)

        self._val, decision = pxu.compute(self._val, decision)
        return decision

    def info(self) -> cabc.Mapping[str, float]:
        if self._val.size == 1:
            data = {f"AbsError[{self._var}]": float(self._val.max())}  # takes the only element available.
        else:
            data = {
                f"AbsError[{self._var}]_min": float(self._val.min()),
                f"AbsError[{self._var}]_max": float(self._val.max()),
            }
        return data

    def clear(self):
        self._val = np.r_[0]


class RelError(pxa.StoppingCriterion):
    """
    Stop iterative solver after relative norm change of a variable (or function thereof) reaches threshold.
    """

    def __init__(
        self,
        eps: pxt.Real,
        var: pxt.VarName = "x",
        f: SVFunction = None,
        norm: pxt.Real = 2,
        satisfy_all: bool = True,
    ):
        """
        Parameters
        ----------
        eps: Real
            Positive threshold.
        var: VarName
            Variable in :py:attr:`pyxu.abc.Solver._mstate` to query.
        f: ~collections.abc.Callable
            Optional function to pre-apply to ``_mstate[var]`` before applying the norm.  Defaults to the identity
            function. The callable should either:

            * accept a scalar input -> output a scalar, or
            * accept an NDArray input -> output an NDArray, i.e same semantics as :py:meth:`~pyxu.abc.Map.apply`.
        norm: Integer, Real
            Ln norm to use >= 0. (Default: L2.)
        satisfy_all: bool
            If True (default) and ``_mstate[var]`` is multi-dimensional, stop if all evaluation points lie below
            threshold.
        """
        try:
            assert eps > 0
            self._eps = eps
        except Exception:
            raise ValueError(f"eps: expected positive threshold, got {eps}.")

        self._var = var
        self._f = f if (f is not None) else (lambda _: _)

        try:
            assert norm >= 0
            self._norm = norm
        except Exception:
            raise ValueError(f"norm: expected non-negative, got {norm}.")

        self._satisfy_all = satisfy_all
        self._val = np.r_[0]  # last computed Ln rel-norm(s) in stop().
        self._x_prev = None  # buffered var from last query.

    def stop(self, state: cabc.Mapping) -> bool:
        x = state[self._var]
        if isinstance(x, pxt.Real):
            x = np.r_[x]
        xp = pxu.get_array_module(x)

        if self._x_prev is None:
            self._x_prev = x.copy()
            # force 1st .info() call to have same format as further calls.
            self._val = np.zeros(shape=(1,) if (x.ndim == 1) else x.shape[:-1])
            return False  # decision deferred: insufficient history to evaluate rel-err.
        else:
            norm = lambda _: xp.linalg.norm(_, ord=self._norm, axis=-1, keepdims=True)
            rule = xp.all if self._satisfy_all else xp.any

            fx_prev = self._f(self._x_prev)
            numerator = norm(self._f(x) - fx_prev)
            denominator = norm(fx_prev)
            decision = rule(numerator <= self._eps * denominator)

            with warnings.catch_warnings():
                # Store relative improvement values for info(). Special care must be taken for the
                # problematic case 0/0 -> NaN.
                warnings.simplefilter("ignore")
                self._val = numerator / denominator
                self._val[xp.isnan(self._val)] = 0  # no relative improvement.
            self._x_prev = x.copy()

            self._x_prev, self._val, decision = pxu.compute(self._x_prev, self._val, decision)
            return decision

    def info(self) -> cabc.Mapping[str, float]:
        if self._val.size == 1:
            data = {f"RelError[{self._var}]": float(self._val.max())}  # takes the only element available.
        else:
            data = {
                f"RelError[{self._var}]_min": float(self._val.min()),
                f"RelError[{self._var}]_max": float(self._val.max()),
            }
        return data

    def clear(self):
        self._val = np.r_[0]
        self._x_prev = None



import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


# We disable PrecisionWarnings since ExplicitLinOp() is not precision-agnostic, but the outputs
# computed must still be valid.
@pytest.mark.filterwarnings("ignore::pyxu.info.warning.PrecisionWarning")
class ExplicitOpMixin:
    # Mixin class which must be inherited from by each concrete ExplicitLinOp sub-class.
    # Reason: sets all parameters such that users only need to provide the linear operator (in array
    # form) being tested: `matrix`.

    # Internal Helpers --------------------------------------------------------
    @staticmethod
    def spec_data() -> list[tuple[callable, pxd.NDArrayInfo, pxrt.Width]]:
        N = pxd.NDArrayInfo
        S = pxd.SparseArrayInfo
        data = []  # (array_initializer, *accepted input backend/width)

        # NUMPY inputs ------------------------
        data.extend(
            itertools.product(
                [
                    N.NUMPY.module().array,
                    S.SCIPY_SPARSE.module().bsr_matrix,
                    S.SCIPY_SPARSE.module().coo_matrix,
                    S.SCIPY_SPARSE.module().csc_matrix,
                    S.SCIPY_SPARSE.module().csr_matrix,
                    S.PYDATA_SPARSE.module().COO.from_numpy,
                    S.PYDATA_SPARSE.module().GCXS.from_numpy,
                ],
                (N.NUMPY,),
                pxrt.Width,
            )
        )

        # DASK inputs -------------------------
        data.extend(
            itertools.product(
                (N.DASK.module().array,),
                (N.DASK,),
                pxrt.Width,
            )
        )

        # CUPY inputs -------------------------
        if pxd.CUPY_ENABLED:
            cp_t = N.CUPY.module().array
            data.extend(
                itertools.product(
                    [
                        cp_t,
                        lambda _: S.CUPY_SPARSE.module().coo_matrix(cp_t(_)),
                        lambda _: S.CUPY_SPARSE.module().csc_matrix(cp_t(_)),
                        lambda _: S.CUPY_SPARSE.module().csr_matrix(cp_t(_)),
                    ],
                    (N.CUPY,),
                    pxrt.Width,
                )
            )

        return data

    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def matrix(self) -> np.ndarray:
        raise NotImplementedError

    @pytest.fixture(params=spec_data())
    def _spec(self, matrix, request):
        init, ndi, width = request.param
        A = init(matrix).astype(width.value)
        op = self.base.from_array(A=A)
        return A, (op, ndi, width)

    @pytest.fixture
    def raw_init_input(self, _spec):
        return _spec[0]

    @pytest.fixture
    def spec(self, _spec):
        return _spec[1]

    @pytest.fixture
    def data_shape(self, matrix):
        return matrix.shape

    @pytest.fixture
    def data_apply(self, matrix):
        N = matrix.shape[1]
        arr = self._random_array((N,))
        out = matrix @ arr
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    # Tests -------------------------------------------------------------------
    def test_array_not_modified(self, op, raw_init_input):
        # Ensure ExplicitLinOp is using the raw array provided to __init__().
        # The user is expected to know what he is doing.
        assert op.mat is raw_init_input


class TestExplicitLinOp(ExplicitOpMixin, conftest.LinOpT):
    @pytest.fixture
    def matrix(self):
        import pyxu_tests.operator.examples.test_linop as tc

        return tc.Tile(N=10, M=5).asarray()


class TestExplicitSquareOp(ExplicitOpMixin, conftest.SquareOpT):
    @pytest.fixture
    def matrix(self):
        import pyxu_tests.operator.examples.test_squareop as tc

        return tc.CumSum(N=19).asarray()


class TestExplicitProjOp(ExplicitOpMixin, conftest.ProjOpT):
    @pytest.fixture
    def matrix(self):
        import pyxu_tests.operator.examples.test_projop as tc

        return tc.Oblique(N=19, alpha=np.pi / 4).asarray()


class TestExplicitOrthProjOp(ExplicitOpMixin, conftest.OrthProjOpT):
    @pytest.fixture
    def matrix(self):
        import pyxu_tests.operator.examples.test_orthprojop as tc

        return tc.ScaleDown(N=19).asarray()


class TestExplicitNormalOp(ExplicitOpMixin, conftest.NormalOpT):
    @pytest.fixture
    def matrix(self):
        import pyxu_tests.operator.examples.test_normalop as tc

        return tc.CircularConvolution(h=np.ones(20)).asarray()


class TestExplicitUnitOp(ExplicitOpMixin, conftest.UnitOpT):
    @pytest.fixture
    def matrix(self):
        import pyxu_tests.operator.examples.test_unitop as tc

        return tc.Permutation(N=19).asarray()


class TestExplicitSelfAdjointOp(ExplicitOpMixin, conftest.SelfAdjointOpT):
    @pytest.fixture
    def matrix(self):
        import pyxu_tests.operator.examples.test_selfadjointop as tc

        return tc.SelfAdjointConvolution(N=19).asarray()


class TestExplicitPosDefOp(ExplicitOpMixin, conftest.PosDefOpT):
    @pytest.fixture
    def matrix(self):
        import pyxu_tests.operator.examples.test_posdefop as tc

        return tc.PSDConvolution(N=19).asarray()


class TestExplicitLinFunc(ExplicitOpMixin, conftest.LinFuncT):
    @pytest.fixture
    def matrix(self):
        import pyxu_tests.operator.examples.test_linfunc as tc

        return tc.ScaledSum(N=19).asarray()



# This stub file should contain the test suite of pyxu.operator.interop.source.from_source().
#
# from_source() is a low-level function used throughout the Pyxu API; as such it is indirectly
# tested through concrete instantiations in:
#    pyxu.operator.linop.base.py
#    pyxu.abc.arithmetic.py
#    (and many others)
#
# Should specific tests be added in the future, they will be inserted here.



import collections.abc as cabc
import typing as typ

import pytest

import pyxu.experimental.sampler as pxe_sampler
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu_tests.conftest as ct


class SamplerT(ct.DisableTestMixin):
    # Internal helpers --------------------------------------------------------
    @classmethod
    def _check_shape(
        cls,
        samples_list,
    ):
        assert all([s.shape == samples_list[0].shape for s in samples_list[1:]])

    @staticmethod
    def _check_backend(samples_list):
        assert all([isinstance(s, type(samples_list[0])) for s in samples_list[1:]])

    @staticmethod
    def _check_prec(samples_list):
        assert all([s.dtype == samples_list[0].dtype for s in samples_list[1:]])

    @staticmethod
    def _check_precCM(
        gen,
        widths: cabc.Collection[pxrt.Width] = pxrt.Width,
    ):
        stats = dict()
        for w in widths:
            with pxrt.Precision(w):
                out = next(gen)
            stats[w] = out.dtype == w.value
        assert all(stats.values())

    @staticmethod
    def _check_reproducibility(samples_1, samples_2, num_samples, xp):
        with pxrt.EnforcePrecision(False):
            for _ in range(num_samples):
                sample_1 = next(samples_1)
                sample_2 = next(samples_2)
                assert xp.allclose(sample_1, sample_2)

    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def spec(self) -> tuple[typ.Generator, pxd.NDArrayInfo, pxrt.Width]:
        # override in subclass to return:
        # * the sampler.samples() generator, where sampler is the sampler to test;
        # * the backend of accepted input arrays;
        # * the precision of accepted input arrays.
        #
        # The triplet (sampler, backend, precision) must be provided since some samplers may not be
        # backend/precision-agnostic.
        raise NotImplementedError

    @pytest.fixture
    def samples(self, spec) -> typ.Generator:
        return spec[0]

    @pytest.fixture
    def samples_copy(self) -> typ.Generator:
        # override in subclass to return a copy of the sampler.samples() generator (with the same rng)
        raise NotImplementedError

    @pytest.fixture
    def sampler(self) -> pxe_sampler._Sampler:
        # override in subclass to return the sampler object being tested.
        raise NotImplementedError

    @pytest.fixture
    def seed(self) -> pxrt.Width:
        return 1234

    @pytest.fixture
    def num_samples(self) -> int:
        return 5

    @pytest.fixture
    def samples_list(self, num_samples, samples) -> list[pxt.NDArray]:
        with pxrt.EnforcePrecision(False):
            s_list = [next(samples) for _ in range(num_samples)]
        return s_list

    @pytest.fixture
    def ndi(self, spec) -> pxd.NDArrayInfo:
        ndi_ = spec[1]
        if ndi_.module() is None:
            pytest.skip(f"{ndi_} unsupported on this machine.")
        return ndi_

    @pytest.fixture
    def xp(self, ndi) -> pxt.ArrayModule:
        return ndi.module()

    @pytest.fixture
    def width(self, spec) -> pxrt.Width:
        return spec[2]

    # Tests -------------------------------------------------------------------
    def test_shape_samples(self, samples_list):
        self._skip_if_disabled()
        self._check_shape(samples_list)

    def test_backend_samples(self, samples_list):
        self._skip_if_disabled()
        self._check_backend(samples_list)

    def test_prec_samples(self, samples_list):
        self._skip_if_disabled()
        self._check_prec(samples_list)

    def test_precCM_samples(self, samples):
        self._skip_if_disabled()
        self._check_precCM(samples)

    def test_reproducibility_samples(self, samples, samples_copy, num_samples, xp):
        self._skip_if_disabled()
        self._check_reproducibility(samples, samples_copy, num_samples, xp)



import numpy as np

import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.operator.interop.source as px_src
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "ConstantValued",
]


def ConstantValued(
    shape: pxt.OpShape,
    cst: pxt.Real,
) -> pxt.OpT:
    r"""
    Constant-valued operator :math:`C: \mathbb{R}^{M} \to \mathbb{R}^{N}`.
    """
    cst = float(cst)
    if np.isclose(cst, 0):
        from pyxu.operator import NullOp

        op = NullOp(shape=shape)
    else:

        @pxrt.enforce_precision(i="arr")
        def op_apply(_, arr: pxt.NDArray) -> pxt.NDArray:
            xp = pxu.get_array_module(arr)
            x = xp.full((1,), fill_value=_._cst, dtype=arr.dtype)
            out = xp.broadcast_to(x, (*arr.shape[:-1], _.codim))
            return out

        def op_jacobian(_, arr: pxt.NDArray) -> pxt.OpT:
            from pyxu.operator import NullOp

            return NullOp(shape=_.shape).squeeze()

        @pxrt.enforce_precision(i="arr")
        def op_grad(_, arr: pxt.NDArray) -> pxt.NDArray:
            xp = pxu.get_array_module(arr)
            x = xp.zeros((1,), dtype=arr.dtype)
            out = xp.broadcast_to(x, arr.shape)
            return out

        @pxrt.enforce_precision(i=("arr", "tau"))
        def op_prox(_, arr: pxt.NDArray, tau: pxt.NDArray) -> pxt.NDArray:
            return pxu.read_only(arr)

        op = px_src.from_source(
            cls=pxa.ProxDiffFunc if (shape[0] == 1) else pxa.DiffMap,
            shape=shape,
            embed=dict(
                _name="ConstantValued",
                _cst=cst,
            ),
            apply=op_apply,
            jacobian=op_jacobian,
            grad=op_grad,
            prox=op_prox,
        )
        op.lipschitz = 0
        op.diff_lipschitz = 0
    return op.squeeze()



import collections.abc as cabc

import numpy as np

import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.operator.interop.source as px_src
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "Sum",
]


def Sum(
    arg_shape: pxt.NDArrayShape,
    axis: pxt.NDArrayAxis = None,
) -> pxt.OpT:
    r"""
    Multi-dimensional sum reduction.

    This operator re-arranges the input array to a multi-dimensional array of shape `arg_shape`, then reduces it via
    summation across one or more `axis`.

    For example, assuming the input array :math:`\mathbf{x} \in \mathbb{R}^{N_1 \times N_2 \times N_3}` and ``axis=-1``,
    then

    .. math::

       \mathbf{y}_{i,j} = \sum_{k}{\mathbf{x}_{i,j,k}}.

    Parameters
    ----------
    arg_shape: NDArrayShape
        Shape of the data to be reduced.
    axis: NDArrayAxis
        Axis or axes along which a sum is performed.  The default, axis=None, will sum all the elements of the input
        array.  If axis is negative it counts from the last to the first axis.

    Notes
    -----
    The matrix operator of a 1D reduction applied to :math:`\mathbf{x} \in \mathbb{R}^{N}` is given by

    .. math::

       \mathbf{A}(x) = \mathbf{1}^{T} \mathbf{x},

    where :math:`\sigma_{\max}(\mathbf{A}) = \sqrt{N}`.  An ND reduction is a chain of 1D reductions in orthogonal
    dimensions.  Hence the Lipschitz constant of an ND reduction is the product of Lipschitz constants of all 1D
    reductions involved, i.e.:

    .. math::

       L = \sqrt{\prod_{i_{k}} N_{i_{k}}},

    where :math:`\{i_{k}\}_{k}` denotes the axes being summed over.
    """

    def as_array(obj) -> np.ndarray:
        if isinstance(obj, cabc.Sequence):
            pass
        else:
            obj = [obj]
        return np.array(obj, dtype=int)

    arg_shape = as_array(arg_shape)
    assert np.all(arg_shape > 0)
    N_dim = len(arg_shape)

    if axis is None:
        axis = np.arange(N_dim)
    axis = np.unique(as_array(axis))  # drop potential duplicates
    assert np.all((-N_dim <= axis) & (axis < N_dim))  # all axes in valid range
    axis = (axis + N_dim) % N_dim  # get rid of negative axes

    sum_shape = arg_shape.copy()  # array shape after reduction
    sum_shape[axis] = 1

    @pxrt.enforce_precision(i="arr")
    def op_apply(_, arr: pxt.NDArray) -> pxt.NDArray:
        sh = arr.shape[:-1]
        arr = arr.reshape(sh + _._arg_shape)

        axis = tuple(ax + len(sh) for ax in _._axis)
        out = arr.sum(axis=axis).reshape(*sh, -1)

        return out

    @pxrt.enforce_precision(i="arr")
    def op_adjoint(_, arr: pxt.NDArray) -> pxt.NDArray:
        sh = arr.shape[:-1]
        arr = arr.reshape(sh + _._sum_shape)

        xp = pxu.get_array_module(arr)
        out = xp.broadcast_to(arr, sh + _._arg_shape)

        out = out.reshape(*sh, -1)
        return out

    def op_estimate_lipschitz(_, **kwargs) -> pxt.Real:
        N = np.prod(_._arg_shape) / np.prod(_._sum_shape)
        L = np.sqrt(N)
        return L

    dim = arg_shape.prod()
    codim = dim // arg_shape[axis].prod()

    op = px_src.from_source(
        cls=pxa.LinOp if codim > 1 else pxa.LinFunc,
        shape=(codim, dim),
        embed=dict(
            _name="Sum",
            _axis=tuple(axis),
            _arg_shape=tuple(arg_shape),
            _sum_shape=tuple(sum_shape),
        ),
        apply=op_apply,
        adjoint=op_adjoint,
        estimate_lipschitz=op_estimate_lipschitz,
    )
    op.lipschitz = op.estimate_lipschitz()
    return op



import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.operator.conftest as conftest
import pyxu_tests.operator.linop.fft.conftest_nufft as conftest_nufft


class TestNUFFT3(conftest_nufft.NUFFT_Mixin, conftest.LinOpT):
    # (Extra) Fixtures which parametrize operator -----------------------------
    @pytest.fixture(params=[10, 13])
    def transform_x(self, transform_dimension, request) -> np.ndarray:
        # (M, D) D-dimensional sample points :math:`\mathbf{x}_{j} \in \mathbb{R}^{D}`.
        rng = np.random.default_rng(0)
        x = rng.normal(size=(request.param, transform_dimension))
        return x

    @pytest.fixture(params=[11, 22])
    def transform_z(self, transform_dimension, request) -> np.ndarray:
        # (N, D) D-dimensional query points :math:`\mathbf{z}_{k} \in \mathbb{R}^{D}`.
        rng = np.random.default_rng(1)
        x = rng.normal(size=(request.param, transform_dimension))
        return x

    @pytest.fixture
    def _transform_cArray(
        self,
        transform_x,
        transform_z,
        transform_sign,
    ) -> np.ndarray:
        # Ground-truth LinOp A: \bC^{M} -> \bC^{N} which encodes the type-3 transform.
        A = np.exp(1j * transform_sign * transform_z @ transform_x.T)  # (N, M)
        return A

    # Fixtures from conftest.LinOpT -------------------------------------------
    @pytest.fixture(
        params=itertools.product(
            [
                pxd.NDArrayInfo.NUMPY,
                pxd.NDArrayInfo.DASK,
            ],
            pxrt.Width,
        )
    )
    def spec(
        self,
        transform_x,
        transform_z,
        transform_sign,
        transform_eps,
        transform_real,
        transform_ntrans,
        transform_nthreads,
        transform_modeord,
        request,
    ):
        ndi, width = request.param
        xp, dtype = ndi.module(), width.value
        with pxrt.Precision(width):
            op = pxo.NUFFT.type3(
                x=xp.array(transform_x, dtype=dtype),
                z=xp.array(transform_z, dtype=dtype),
                isign=transform_sign,
                eps=transform_eps,
                real=transform_real,
                enable_warnings=False,
                n_trans=transform_ntrans,
                nthreads=transform_nthreads,
                modeord=transform_modeord,
            )
        return op, ndi, width

    @pytest.fixture
    def data_shape(
        self,
        transform_x,
        transform_z,
        transform_real,
    ):
        dim = len(transform_x) * (1 if transform_real else 2)
        codim = 2 * len(transform_z)
        return (codim, dim)

    @pytest.fixture
    def data_apply(
        self,
        _transform_cArray,
        transform_real,
    ):
        N, M = _transform_cArray.shape

        rng = np.random.default_rng(3)
        w = rng.normal(size=(M,)) + 1j * rng.normal(size=(M,))
        if transform_real:
            w = w.real
        v = _transform_cArray @ w

        return dict(
            in_=dict(arr=w if transform_real else pxu.view_as_real(w)),
            out=pxu.view_as_real(v),
        )


class TestNUFFT3Chunked(TestNUFFT3):
    # (Extra) Fixtures which parametrize operator -----------------------------
    @pytest.fixture(params=[True, False])
    def transform_parallel(self, request) -> bool:
        return request.param

    # Fixtures from conftest.LinOpT -------------------------------------------
    @pytest.fixture(
        params=itertools.product(
            [
                pxd.NDArrayInfo.NUMPY,
                pxd.NDArrayInfo.DASK,
            ],
            pxrt.Width,
        )
    )
    def spec(
        self,
        transform_x,
        transform_z,
        transform_sign,
        transform_eps,
        transform_real,
        transform_ntrans,
        transform_nthreads,
        transform_modeord,
        transform_parallel,
        request,
    ):
        ndi, width = request.param
        xp, dtype = ndi.module(), width.value
        with pxrt.Precision(width):
            op = pxo.NUFFT.type3(
                x=xp.array(transform_x, dtype=dtype),
                z=xp.array(transform_z, dtype=dtype),
                isign=transform_sign,
                eps=transform_eps,
                real=transform_real,
                enable_warnings=False,
                n_trans=transform_ntrans,
                nthreads=transform_nthreads,
                modeord=transform_modeord,
                chunked=True,
                parallel=transform_parallel,
            )

            # Extra initialization steps for chunked transforms
            x_chunks, z_chunks = op.auto_chunk()
            op.allocate(x_chunks, z_chunks)
        return op, ndi, width



import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.operator.conftest as conftest


def filterF(N: int) -> np.ndarray:
    hF = np.r_[1.0, np.r_[1 : (N - 1) // 2 + 1], -np.r_[-(N // 2) : 0]]
    hF /= np.abs(hF).max()
    return hF


class PSDConvolution(pxa.PosDefOp):
    # Convolution where the filter coefficients imply operator is positive-definite.
    def __init__(self, N: int):
        assert N % 2 == 1, "Even-length filters are unsupported."
        super().__init__(shape=(N, N))
        self.lipschitz = np.inf
        self._hF = filterF(N)

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr):
        xp = pxu.get_array_module(arr)
        fw = lambda _: xp.fft.fft(_, axis=-1)
        bw = lambda _: xp.fft.ifft(_, axis=-1)
        hF = xp.array(self._hF, dtype=arr.dtype)
        out = bw(hF * fw(arr)).real
        return out.astype(arr.dtype, copy=False)


class TestPSDConvolution(conftest.PosDefOpT):
    @pytest.fixture(
        params=itertools.product(
            ((11, PSDConvolution(N=11)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (dim, dim)

    @pytest.fixture
    def data_apply(self, dim):
        F = np.fft.ifft(filterF(dim)).real
        N = F.size
        arr = self._random_array((N,))
        out = np.zeros((N,))
        for n in range(N):
            for k in range(N):
                out[n] += arr[k] * F[n - k % N]
        return dict(
            in_=dict(arr=arr),
            out=out,
        )



import numpy as np
import pytest

import pyxu.runtime as pxrt


class TestPrecisionContextManager:
    @pytest.mark.parametrize("w", pxrt.Width)
    def test_contextManager(self, w: pxrt.Width):
        with pxrt.Precision(w):
            assert pxrt.getPrecision() == w


class TestEnforcePrecisionContextManager:
    @pytest.mark.parametrize("state", [True, False])
    def test_contextManager(self, state: bool):
        assert pxrt.getCoerceState() is True  # default
        with pxrt.EnforcePrecision(state):
            assert pxrt.getCoerceState() == state


class TestEnforcePrecisionDecorator:
    def right_dtype(self, x):
        return x.dtype == pxrt.getPrecision().value

    def test_CM_noOp(self):
        @pxrt.enforce_precision("x")
        def f(x):
            return x

        with pxrt.EnforcePrecision(False):
            x = 1
            assert x is f(x)

    @pytest.mark.parametrize(
        "value",
        [
            0.1,  # Python float
            1,  # Python int
            np.float128(1),  # NumPy dtypes
            np.int8(-1),
            np.uint8(1),
            True,  # Python T/F
            False,
        ],
    )
    def test_valid_scalar_io(self, value):
        @pxrt.enforce_precision("x")
        def f(x):
            assert self.right_dtype(x)
            return x + 1

        with pxrt.Precision(pxrt.Width.SINGLE):
            assert self.right_dtype(f(value))

    @pytest.mark.parametrize(
        "value",
        [
            1j,  # Python complex
            np.complex64(2),  # NumPy complex dtype
        ],
    )
    def test_invalid_scalar_io(self, value):
        @pxrt.enforce_precision("xx")
        def f(xx):  # explicitly test multi-character strings
            assert self.right_dtype(xx)
            return xx + 1

        with pytest.raises(TypeError):
            f(value)

    def test_multi_i(self):
        @pxrt.enforce_precision(["x", "y"])
        def f(x, y, z):
            assert self.right_dtype(x)
            assert self.right_dtype(y)
            return x + y

        with pxrt.Precision(pxrt.Width.SINGLE):
            f(1, 1.0, True)

    def test_valid_name_i(self):
        @pxrt.enforce_precision("y")
        def f(x):
            return x

        with pytest.raises(ValueError):
            f(1)

    def test_valid_array_io(self, xp):
        @pxrt.enforce_precision("x")
        def f(x):
            assert self.right_dtype(x)
            return x

        with pxrt.Precision(pxrt.Width.SINGLE):
            x = xp.arange(5)
            assert self.right_dtype(f(x))

    def test_invalid_array_io(self, xp):
        @pxrt.enforce_precision("x")
        def f(x):
            assert self.right_dtype(x)
            return x

        with pxrt.Precision(pxrt.Width.SINGLE):
            x = xp.arange(5) + 1j * xp.arange(5)
            with pytest.raises(TypeError):
                f(x)

    @pytest.mark.parametrize("allow_None", [True, False])
    def test_None_input(self, allow_None):
        @pxrt.enforce_precision("x", allow_None=allow_None)
        def f(x):
            return x

        if allow_None:
            assert f(None) is None
        else:
            with pytest.raises(ValueError):
                f(None)


class TestCoerce:
    def right_dtype(self, x):
        return x.dtype == pxrt.getPrecision().value

    @pytest.mark.parametrize(
        "value",
        [
            0.1,  # Python float
            1,  # Python int
            np.float128(1),  # NumPy dtypes
            np.int8(-1),
            np.uint8(1),
            True,  # Python T/F
            False,
        ],
    )
    def test_valid_scalar(self, value):
        with pxrt.Precision(pxrt.Width.SINGLE):
            assert self.right_dtype(pxrt.coerce(value))

    @pytest.mark.parametrize(
        "value",
        [
            1j,  # Python complex
            np.complex64(2),  # NumPy complex dtype
        ],
    )
    def test_invalid_scalar(self, value):
        with pytest.raises(TypeError):
            pxrt.coerce(value)

    def test_valid_array(self, xp):
        with pxrt.Precision(pxrt.Width.SINGLE):
            x = xp.arange(5)
            assert self.right_dtype(pxrt.coerce(x))

    def test_invalid_array(self, xp):
        with pxrt.Precision(pxrt.Width.SINGLE):
            x = xp.arange(5) + 1j * xp.arange(5)
            with pytest.raises(TypeError):
                pxrt.coerce(x)



import warnings
from importlib.metadata import entry_points

import pyxu.info.warning as pxw


def _load_entry_points(glob, group, names=None):
    r"""
    Load (if any) entry point Pyxu plugins.

    Pyxu accepts contributions in the form of Pyxu-plugins, which can be developed by third-party contributors and are
    not necessarily tested or verified by the team of Pyxu core developers. While we strive to ensure the safety and
    security of our software framework and its plugins, we cannot guarantee their functionality or safety.  Users should
    exercise caution when installing and using plugins and assume full responsibility for any damages or issues that may
    arise from their use. The developers of this software framework are not liable for any harm caused by the use of
    third-party plugins.

    NOTE
    ----
    There might be duplicated extensions when installing a plugin in editable mode (i.e. with pip install -e). This does
    not represent a problem in practice but a warning that an attempt at overloading a Pyxu base object might arise. See
    the issue https://github.com/pypa/setuptools/issues/3649 for further information.

    """
    eps = tuple(entry_points(group=group))

    # Check for duplicated entry points
    seen = set()
    duplicated = [ep.name for ep in eps if ep in seen or seen.add(ep.name)]
    if len(duplicated):
        warnings.warn(f"Found duplicated entry points: {duplicated}.", pxw.ContributionWarning)

    # Load entry points
    try:
        for i, ep in enumerate(eps):
            ep_load = ep.load()
            name = ep.name
            # If plugin can overload, load directly
            if name.startswith("_"):
                if name[1:] in glob:
                    warnings.warn(
                        f"Plugin `{name}` overloaded an existing Pyxu base class/function, use with " f"caution.",
                        pxw.ContributionWarning,
                    )
                    glob[name[1:]] = ep_load
                    if names is not None:
                        names.append(name[1:])

                else:
                    warnings.warn(
                        f"Attempted to overload a non existing Pyxu base class/function `{name}`."
                        f"Do not use the prefix `_`.",
                        pxw.ContributionWarning,
                    )

            # Else, check if class/function already exists in Pyxu
            else:
                if name in glob:
                    warnings.warn(
                        f"Attempting to overload Pyxu base class/function `{name}`.\n"
                        + "Overloading plugins must start with underscore `_`.\n"
                        + f"Defaulting to base class/function `{name}`.\n",
                        pxw.ContributionWarning,
                    )
                else:
                    glob[name] = ep_load
                    if names is not None:
                        names.append(name)
                    warnings.warn(f"Plugin `{name}` loaded, use with caution.", pxw.ContributionWarning)
    except Exception as exc:
        warnings.warn(f"Failed to load plugin `{name}`: {exc}.", pxw.ContributionWarning)

    return names
